@@ -1,19 +1,12 @@
 /*************************************************
-*    The PMW Music Typesetter - 3rd incarnation  *
+*              PMW font functions                *
 *************************************************/
 
-/* Copyright (c) Philip Hazel, 1991 - 2020 */
+/* Copyright Philip Hazel 2021 */
+/* This file created: January 2021 */
+/* This file last modified: July 2022 */
 
-/* Written by Philip Hazel, starting November 1991 */
-/* Re-coded for Unicode support: August 2005 */
-/* This file last modified: November 2020 */
-
-
-/* This file contains code for handling text fonts */
-
-
-#include "pmwhdr.h"
-#include "readhdr.h"
+#include "pmw.h"
 
 
 /* Entries in the character name to Unicode code point table */
@@ -27,7 +20,9 @@ typedef struct an2uencod {
 /* This table translates character names from PostScript fonts that use Adobe's
 standard encoding into Unicode. In addition, for characters whose Unicode
 values are greater than LOWCHARLIMIT, it includes the offset above LOWCHARLIMIT
-that we use for printing these characters. */
+that we use for printing these characters (essentially a private encoding). The
+current highest offset is 43. The value of FONTWIDTHS_SIZE must be large enough
+to accommodate all these characters. */
 
 static an2uencod an2ulist[] = {
   { US"A",               0x0041, -1 },
@@ -434,7 +429,7 @@ while (top > bot)
   if (c > 0) bot = mid + 1; else top = mid;
   }
 
-if (warn) error_moan(ERR122, fname, cname, US" not found");  /* Hard */
+if (warn) error(ERR58, fname, cname, US" not found");  /* Hard */
 return -1;
 }
 
@@ -463,6 +458,7 @@ return (int)(ka->pair - kb->pair);
 }
 
 
+
 /*************************************************
 *         UTR table sorting comparison           *
 *************************************************/
@@ -520,7 +516,7 @@ return p;
 /* This is an externally-called function, also called from below.
 
 Arguments:
-  psname     the font's PostScript name
+  name       the font's name
   ext        ".afm", ".utr", ".pfa", etc., or ""
   fextras    list of extra directories to search
   fdefault   the default directory
@@ -531,7 +527,7 @@ Returns:   the opened file or NULL (when not mandatory)
 */
 
 FILE *
-font_finddata(uschar *psname, const char *ext, uschar *fextras,
+font_finddata(uschar *name, const char *ext, uschar *fextras,
   uschar *fdefault, uschar *filename, BOOL mandatory)
 {
 FILE *f = NULL;
@@ -552,7 +548,7 @@ if (fextras != NULL)
 
     len = ep - pp;
     (void)memcpy(filename, pp, len);
-    sprintf(CS (filename + len), "/%s%s", psname, ext);
+    sprintf(CS (filename + len), "/%s%s", name, ext);
     f = Ufopen(filename, "r");
     if (f != NULL || *ep == 0) break;
     pp = ep + 1;
@@ -563,15 +559,15 @@ if (fextras != NULL)
 
 if (f == NULL)
   {
-  sprintf(CS filename, "%s/%s%s", fdefault, psname, ext);
+  sprintf(CS filename, "%s/%s%s", fdefault, name, ext);
   f = Ufopen(filename, "r");
   if (mandatory && f == NULL)
     {
     const char *s = (*ext == 0)? "Data" : ext;
     if (fextras == NULL)
-      error_moan(ERR59, s, psname, fdefault);            /* Hard error */
+      error(ERR56, s, name, fdefault);           /* Hard error */
     else
-      error_moan(ERR127, s, psname, fextras, fdefault);  /* Hard error */
+      error(ERR57, s, name, fextras, fdefault);  /* Hard error */
     }
   }
 
@@ -580,6 +576,37 @@ return f;
 
 
 
+/*************************************************
+*        Check character value in std font       *
+*************************************************/
+
+/* In a standardly encoded font, characters less than LOWCHARLIMIT are encoded
+as their Unicode values. There are also around 40 much higher-valued characters
+that are translated to the code points immediately above LOWCHARLIMIT so that
+they can be accessed from the second encoding of the font. This function checks
+and translates characters when necessary. It is called when reading kerning
+tables for a standardly-encoded font. Just ignore anything unrecognized.
+
+Arguments:
+  c         the code point
+  fs        the font structure
+
+Returns:    possibly changed character
+*/
+
+static uint32_t
+check_lowchar(uint32_t c, fontstr *fs)
+{
+tree_node *t;
+uschar utf[8];
+if (c < LOWCHARLIMIT || (fs->flags & ff_stdencoding) == 0) return c;
+utf[misc_ord2utf8(c, utf)] = 0;
+t = tree_search(fs->high_tree, utf);
+return (t == NULL)? c : (uint32_t)(LOWCHARLIMIT + t->value);
+}
+
+
+
 /*************************************************
 *  Load width, kern, & Unicode tables for a font *
 *************************************************/
@@ -588,49 +615,234 @@ return f;
 containing character widths and kerning information, and also checks for an
 optional UTR file if the font is not standardly-encoded.
 
-Argument:   the font id (offset in font_List)
+Argument:   the font id (offset in font_list)
 Returns:    nothing
 */
 
 void
-font_loadtables(int fontid)
+font_loadtables(uint32_t fontid)
 {
-FILE *f;
+FILE *fa, *fu;
 int kerncount = 0;
 int finalcount = 0;
-int *widths, *r2ladjusts, *heights;
 kerntablestr *kerntable;
-fontstr *fs = &(font_List[fontid]);
+fontstr *fs = &(font_list[fontid]);
+tree_node *treebase = NULL;
 uschar *pp;
 uschar filename[256];
 uschar line[256];
 
-DEBUG(("Loading AFM for %s\n", fs->psname));
+/* Find and open the AFM file */
 
-f = font_finddata(fs->psname, ".afm", font_data_extra, font_data_default,
+fa = font_finddata(fs->name, ".afm", font_data_extra, font_data_default,
   filename, TRUE);   /* Hard error if not found */
-widths = fs->widths = malloc(LOWCHARLIMIT * sizeof(int));
-memset(widths, 0, LOWCHARLIMIT * sizeof(int));
-r2ladjusts = fs->r2ladjusts = malloc(LOWCHARLIMIT * sizeof(int));
-memset(r2ladjusts, 0, LOWCHARLIMIT * sizeof(int));
-heights = fs->heights = NULL;
+
+fs->widths = mem_get_independent(FONTWIDTHS_SIZE * sizeof(int32_t));
+memset(fs->widths, 0, FONTWIDTHS_SIZE * sizeof(int32_t));
+fs->r2ladjusts = mem_get_independent(FONTWIDTHS_SIZE * sizeof(int32_t));
+memset(fs->r2ladjusts, 0, FONTWIDTHS_SIZE * sizeof(int32_t));
+fs->heights = NULL;
 fs->kerncount = 0;
 
-/* Process the AFM file. First find the start of the metrics; on the way, check
-for the standard encoding scheme and for fixed pitch. */
+/* Find the start of the metrics in the AFM file; on the way, check for the
+standard encoding scheme and for fixed pitch. */
 
 for (;;)
   {
-  if (Ufgets(line, sizeof(line), f) == NULL)
-    error_moan(ERR122, filename, "no metric data found", "");  /* Hard */
+  if (Ufgets(line, sizeof(line), fa) == NULL)
+    error(ERR58, filename, "no metric data found", "");  /* Hard */
   if (memcmp(line, "EncodingScheme AdobeStandardEncoding", 36) == 0)
-    fs->stdencoding = TRUE;
+    fs->flags |= ff_stdencoding;
   if (memcmp(line, "IsFixedPitch true", 17) == 0)
-    fs->fixedpitch = TRUE;
+    fs->flags |= ff_fixedpitch;
   if (memcmp(line, "StartCharMetrics", 16) == 0) break;
   }
 
-/* Process the metric lines for each character */
+/* Now we know whether or not the font has standard encoding we can set up the
+unsupported character substitution, and look for a .utr file with Unicode
+translations (and maybe an unsupported character) for fonts that are not
+standardly encoded. */
+
+if ((fs->flags & ff_stdencoding) == 0)
+  {
+  fs->invalid = UNKNOWN_CHAR_N | (font_mf << 24);  /* Default */
+  fu = font_finddata(fs->name, ".utr", font_data_extra, font_data_default,
+    filename, FALSE);
+  }
+else
+  {
+  fs->invalid = UNKNOWN_CHAR_S | (font_unknown << 24);
+  fu = NULL;
+  }
+
+/* Process a .utr file */
+
+if (fu != NULL)
+  {
+  int ucount = 0;
+  int lineno = 0;
+  uint32_t limit_pscode = 256;
+  utrtablestr utable[MAX_UTRANSLATE];
+
+  TRACE("Loading UTR for %s\n", fs->name);
+
+  while (Ufgets(line, sizeof(line), fu) != NULL)
+    {
+    uschar *epp;
+
+    lineno++;
+    pp = line;
+    while (isspace(*pp)) pp++;
+    if (*pp == 0 || *pp == '#') continue;
+
+    /* A line beginning with '/' is an encoding definition line. */
+
+    if (*pp == '/')
+      {
+      tree_node *tn;
+      uschar *nb, *ne;
+      uint32_t c;
+
+      if (fs->encoding == NULL)
+        {
+        fs->encoding = mem_get_independent(FONTWIDTHS_SIZE * sizeof(char *));
+        for (int i = 0; i < FONTWIDTHS_SIZE; i++) fs->encoding[i] = NULL;
+        limit_pscode = FONTWIDTHS_SIZE;
+        }
+
+      nb = ++pp;
+      while (isalnum(*pp)) pp++;
+      ne = pp;
+
+      while (isspace(*pp)) pp++;
+      c = (uint32_t)Ustrtoul(pp, &epp, 0);
+      if (epp == pp)
+        {
+        error(ERR60, "font", lineno, filename, line);
+        continue;
+        }
+      if (c >= FONTWIDTHS_SIZE)
+        {
+        error(ERR61, FONTWIDTHS_SIZE, lineno, filename, line);
+        continue;
+        }
+      if (fs->encoding[c] != NULL)
+        {
+        error(ERR176, c, lineno, filename, line);
+        continue;
+        }
+
+      /* We now have a name and a code. The name is saved and pointed to from
+      the encoding vector for appropriate PostScript generation. We also create
+      a tree of names so that the name and its code can be quickly found when
+      reading the AFM file below. */
+
+      *ne = 0;
+      tn = mem_get(sizeof(tree_node));
+      tn->name = fs->encoding[c] = mem_copystring(nb);
+      tn->value = c;
+      if (!tree_insert(&treebase, tn))
+        error(ERR175, tn->name, lineno, filename, line);
+      continue;
+      }
+
+    /* Handle an "unknown character" definition */
+
+    if (*pp == '?')
+      {
+      uint32_t c = (uint32_t)Ustrtoul(++pp, &epp, 0);
+      if (epp == pp)
+        {
+        error(ERR60, "font", lineno, filename, line);
+        continue;
+        }
+      if (c >= limit_pscode)
+        {
+        error(ERR61, limit_pscode, lineno, filename, line);
+        continue;
+        }
+
+      /* Change an underlay special character to its escaped version. */
+
+      switch (c)
+        {
+        case '#': c = ss_escapedsharp; break;
+        case '=': c = ss_escapedequals; break;
+        case '-': c = ss_escapedhyphen; break;
+        default: break;
+        }
+
+      fs->invalid = c | (font_unknown << 24);
+      continue;
+      }
+
+    /* Handle a translation definition */
+
+    if (ucount >= MAX_UTRANSLATE)
+      {
+      error(ERR59, filename, MAX_UTRANSLATE);
+      break;
+      }
+
+    if (Ustrncmp(pp, "U+", 2) == 0) pp += 2;
+    utable[ucount].unicode = (unsigned int)Ustrtoul(pp, &epp, 16);
+    if (epp == pp)
+      {
+      error(ERR60, "Unicode", lineno, filename, line);
+      continue;
+      }
+    pp = epp;
+    utable[ucount].pscode = (unsigned int)Ustrtoul(pp, &epp, 0);
+    if (epp == pp)
+      {
+      error(ERR60, "font", lineno, filename, line);
+      continue;
+      }
+    if (utable[ucount].pscode >= limit_pscode)
+      {
+      error(ERR61, limit_pscode, lineno, filename, line);
+      continue;
+      }
+
+    /* Change underlay special characters to escaped versions. */
+
+    switch (utable[ucount].pscode)
+      {
+      case '#': utable[ucount].pscode = ss_escapedsharp; break;
+      case '=': utable[ucount].pscode = ss_escapedequals; break;
+      case '-': utable[ucount].pscode = ss_escapedhyphen; break;
+      default: break;
+      }
+
+    ucount++;
+    }
+
+  (void)fclose(fu);
+
+  /* Sort the data by Unicode value, check for duplicates, and remember with
+  the font. */
+
+  if (ucount > 0)
+    {
+    int i;
+    qsort(utable, ucount, sizeof(utrtablestr), utr_table_cmp);
+    for (i = 1; i < ucount; i++)
+      {
+      if (utable[i].unicode == utable[i-1].unicode)
+        {
+        error(ERR62, utable[i].unicode, filename);
+        while(i < ucount - 1 && utable[i].unicode == utable[i+1].unicode) i++;
+        }
+      }
+    fs->utr = mem_get_independent(ucount * sizeof(utrtablestr));
+    memcpy(fs->utr, utable, ucount * sizeof(utrtablestr));
+    fs->utrcount = ucount;
+    }
+  }
+
+/* Now we can process the metric lines in the AFM file. */
+
+TRACE("Loading AFM for %s\n", fs->name);
 
 for (;;)
   {
@@ -639,12 +851,12 @@ for (;;)
   int poffset = -1;
   int r2ladjust = 0;
 
-  if (Ufgets(line, sizeof(line), f) == NULL)
-    error_moan(ERR122, filename, "unexpected end of metric data", "");  /* Hard */
+  if (Ufgets(line, sizeof(line), fa) == NULL)
+    error(ERR58, filename, "unexpected end of metric data", "");  /* Hard */
   if (memcmp(line, "EndCharMetrics", 14) == 0) break;
 
   if (memcmp(line, "C ", 2) != 0)
-    error_moan(ERR122, filename, "unrecognized metric data line: ", line); /* Hard */
+    error(ERR58, filename, "unrecognized metric data line: ", line); /* Hard */
 
   pp = line + 2;
   while (memcmp(pp, "WX", 2) != 0) pp++;
@@ -680,17 +892,55 @@ for (;;)
     }
 
   /* If this is a StandardEncoding font, scan the list of characters so as to
-  get the Unicode value for this character. */
+  get the Unicode value for this character. If the code point is greater than
+  or equal to LOWCHARLIMIT, the offset above LOWCHARLIMIT that is used for
+  these characters is returned.  */
 
-  if (fs->stdencoding)
+  if ((fs->flags & ff_stdencoding) != 0)
     {
     uschar *cname;
     while (memcmp(pp, "N ", 2) != 0) pp++;
     cname = (pp += 2);
     while (*pp != ' ') pp++;
     *pp = 0;
-    code = an2u(cname, fs->psname, TRUE, &poffset);
+    code = an2u(cname, fs->name, TRUE, &poffset);
     if (code < 0) continue;  /* Don't try to store anything! */
+
+    /* Remember that this font has certain characters */
+
+    if (code == CHAR_FI) fs->flags |= ff_hasfi;
+
+    /* If the Unicode code point is not less than LOWCHARLIMIT, remember it and
+    its special offset in a tree so that it can be translated when encountered in
+    a string. Then set the translated value for saving the width. */
+
+    if (code >= LOWCHARLIMIT)
+      {
+      tree_node *tc = mem_get(sizeof(tree_node));
+      tc->name = mem_get(8);
+      tc->name[misc_ord2utf8(code, tc->name)] = 0;
+      tc->value = poffset;
+      (void)tree_insert(&(fs->high_tree), tc);
+      code = LOWCHARLIMIT + poffset;
+      }
+    }
+
+  /* If a non-standardly encoded font has a .utr file, it will have been read
+  above, and if there is encoding information in it, a tree of character names
+  will exist that translates from glyph name to encoding value. Ignore unknown
+  names. */
+
+  else if (treebase != NULL)
+    {
+    tree_node *tn;
+    uschar *cname;
+    while (memcmp(pp, "N ", 2) != 0) pp++;
+    cname = (pp += 2);
+    while (*pp != ' ') pp++;
+    *pp = 0;
+    tn = tree_search(treebase, cname);
+    if (tn == NULL) continue;
+    code = tn->value;
     }
 
   /* For other fonts, just use the character number directly. If there are
@@ -704,39 +954,22 @@ for (;;)
     while (*pp != 0 && memcmp(pp, "WY", 2) != 0) pp++;
     if (*pp != 0)
       {
-      int height;
+      int32_t height;
       pp += 2;
       (void)read_number(&height, pp);
-      if (heights == NULL)
+      if (fs->heights == NULL)
         {
-        heights = fs->heights = malloc(256 * sizeof(int));
-        memset(heights, 0, 256 * sizeof(int));
+        fs->heights = mem_get_independent(256 * sizeof(int32_t));
+        memset(fs->heights, 0, 256 * sizeof(int));
         }
-      heights[code] = height;
+      fs->heights[code] = height;
       }
     }
 
-  /* Remember that this font has certain characters */
+  /* Save the widths */
 
-  if (code == CHAR_FI) fs->hasfi = TRUE;
-
-  /* Now put the widths in an appropriate place. */
-
-  if (code < LOWCHARLIMIT)
-    {
-    widths[code] = width;
-    r2ladjusts[code] = r2ladjust;
-    }
-  else
-    {
-    tree_node *tc = malloc(sizeof(tree_node));
-    tc->name = malloc(8);
-    tc->name[misc_ord2utf8(code, tc->name)] = 0;
-    tc->val[0] = poffset;
-    tc->val[1] = width;
-    tc->val[2] = r2ladjust;
-    (void)Tree_InsertNode(&(fs->high_tree), tc);
-    }
+  fs->widths[code] = width;
+  fs->r2ladjusts[code] = r2ladjust;
   }
 
 /* Process kerning data (if any); when this is done, we are finished with the
@@ -744,20 +977,20 @@ AFM file. */
 
 for (;;)
   {
-  if (Ufgets(line, sizeof(line), f) == NULL) goto ENDKERN;
+  if (Ufgets(line, sizeof(line), fa) == NULL) goto ENDKERN;
   if (memcmp(line, "StartKernPairs", 14) == 0) break;
   }
 
 /* Find size of kern table, and get space for it. In the past, some of Adobe's
 AFM files had a habit of containing a large number of kern pairs with
 zero amount of kern. We leave these out of the table and adjust the count for
-searching, but don't bother to free up the unused store (it isn't a vast
+searching, but don't bother to free up the unused memory (it isn't a vast
 amount). */
 
 pp = line + 14;
 while (*pp != 0 && *pp == ' ') pp++;
 (void)read_number(&kerncount, pp);
-fs->kerns = kerntable = malloc(kerncount*sizeof(kerntablestr));
+fs->kerns = kerntable = mem_get_independent(kerncount*sizeof(kerntablestr));
 
 finalcount = 0;
 while (kerncount--)
@@ -768,8 +1001,8 @@ while (kerncount--)
   int a = -1;
   int b = -1;
 
-  if (Ufgets(line, sizeof(line), f) == NULL)
-    error_moan(ERR122, filename, "unexpected end of kerning data");  /* Hard */
+  if (Ufgets(line, sizeof(line), fa) == NULL)
+    error(ERR58, filename, "unexpected end of kerning data");  /* Hard */
   if (memcmp(line, "EndKernPairs", 12) == 0) break;
 
   /* Skip blank lines */
@@ -787,19 +1020,22 @@ while (kerncount--)
   x = pp;
   while (*pp != 0 && *pp != ' ') pp++;
   *pp++ = 0;
-  a = an2u(x, fs->psname, FALSE, NULL);
+  a = an2u(x, fs->name, FALSE, NULL);
 
   while (*pp != 0 && *pp == ' ') pp++;
   x = pp;
   while (*pp != 0 && *pp != ' ') pp++;
   *pp++ = 0;
-  b = an2u(x, fs->psname, FALSE, NULL);
+  b = an2u(x, fs->name, FALSE, NULL);
 
   /* Read the kern value only if we have found the characters; otherwise ignore
   the kern */
 
   if (a >= 0 && b >= 0)
     {
+    a = check_lowchar(a, fs);
+    b = check_lowchar(b, fs);
+
     kerntable[finalcount].pair = (a << 16) + b;
     while (*pp != 0 && *pp == ' ') pp++;
     if (*pp == '-') { sign = -1; pp++; }
@@ -816,88 +1052,13 @@ qsort(kerntable, fs->kerncount, sizeof(kerntablestr), kern_table_cmp);
 /* Finished with the AFM file */
 
 ENDKERN:
-(void)fclose(f);
-
-
-/* If this font does not have standard encoding, see if there is a .utr file
-that contains Unicode translations. */
-
-if (!fs->stdencoding &&
-    (f = font_finddata(fs->psname, ".utr", font_data_extra, font_data_default,
-      filename, FALSE)) != NULL)
-  {
-  int ucount = 0;
-  int lineno = 0;
-  utrtablestr utable[MAX_UTRANSLATE];
-
-  DEBUG(("Loading UTR for %s\n", fs->psname));
-
-  while (Ufgets(line, sizeof(line), f) != NULL)
-    {
-    uschar *epp;
-
-    lineno++;
-    pp = line;
-    while (isspace(*pp)) pp++;
-    if (*pp == 0 || *pp == '#') continue;
-
-    if (ucount >= MAX_UTRANSLATE)
-      {
-      error_moan(ERR147, filename, MAX_UTRANSLATE);
-      break;
-      }
-
-    if (Ustrncmp(pp, "U+", 2) == 0) pp += 2;
-    utable[ucount].unicode = (unsigned int)Ustrtoul(pp, &epp, 16);
-    if (epp == pp)
-      {
-      error_moan(ERR148, "Unicode", lineno, filename, line);
-      continue;
-      }
-    pp = epp;
-    utable[ucount].pscode = (unsigned int)Ustrtoul(pp, &epp, 0);
-    if (epp == pp)
-      {
-      error_moan(ERR148, "font", lineno, filename, line);
-      continue;
-      }
-    if (utable[ucount].pscode >= 256)
-      {
-      error_moan(ERR149, lineno, filename, line);
-      continue;
-      }
-
-    ucount++;
-    }
-
-  (void)fclose(f);
-
-  /* Sort the data, check for duplicates, and remember with the font. */
-
-  if (ucount > 0)
-    {
-    int i;
-    qsort(utable, ucount, sizeof(utrtablestr), utr_table_cmp);
-    for (i = 1; i < ucount; i++)
-      {
-      if (utable[i].unicode == utable[i-1].unicode)
-        {
-        error_moan(ERR7, utable[i].unicode, filename);
-        while(i < ucount - 1 && utable[i].unicode == utable[i+1].unicode) i++;
-        }
-      }
-    fs->utr = store_Xget(ucount * sizeof(utrtablestr));
-    memcpy(fs->utr, utable, ucount * sizeof(utrtablestr));
-    fs->utrcount = ucount;
-    }
-  }
-
+(void)fclose(fa);
 
 /* Early checking debugging code; retained in the source in case it is ever
 needed again. */
 
 #ifdef NEVER
-debug_printf("FONT %s\n", fs->psname);
+debug_printf("FONT %s\n", fs->name);
   {
   int i;
 
@@ -936,50 +1097,77 @@ debug_printf("FONT %s\n", fs->psname);
 *         Add a font to the base list            *
 *************************************************/
 
-/* This function is called from font_init() below, to add a single basic
-PostScript font to the font_List vector.
+/* This function is called from font_init() below, and also by the textfont and
+musicfont commands, to add a typeface to the list of fonts.
+
+Arguments:
+  name     name of font
+  fontid   offset in font_table (font_rm etc)
+  flags    font flags
 
-Argument:  Name of PostScript font
 Returns:   nothing
 */
 
-static void
-font_addbasefont(uschar *psfont)
+void
+font_addfont(uschar *name, uint32_t fontid, uint32_t flags)
 {
-fontstr *fs = &(font_List[font_basecount]);
-if (font_basecount >= MAX_FONTS) { error_moan(ERR26, MAX_FONTS); return; }
-fs->psname = malloc(Ustrlen(psfont)+1);
-Ustrcpy(fs->psname, psfont);
+fontstr *fs;
+
+if (font_count >= font_list_size)
+  {
+  font_list_size += FONTLIST_CHUNKSIZE;
+  font_list = realloc(font_list, font_list_size * sizeof(fontstr));
+  if (font_list == NULL) error(ERR0, "re-", "font list vector",
+    font_list_size);  /* Hard */
+  }
+
+fs = &(font_list[font_count]);
+fs->name = mem_copystring(name);
 fs->widths = NULL;
 fs->high_tree = NULL;
+fs->utr = NULL;
+fs->encoding = NULL;
+fs->utrcount = 0;
 fs->heights = NULL;
 fs->kerns = NULL;
 fs->kerncount = -1;
-fs->stdencoding = fs->fixedpitch = fs->hasfi = FALSE;
-fs->include = FALSE;
-font_loadtables(font_basecount++);
+fs->flags = flags;
+
+font_table[fontid] = font_count;
+font_loadtables(font_count++);
 }
 
 
 /*************************************************
-*        Set up the base list of fonts           *
+*                Read a font type                *
 *************************************************/
 
-/* Called at start-up to set up the default fonts
-
-Arguments:  none
-Returns:    nothing
+/*
+Argument:  TRUE to not give an error if the word is unrecognised
+Returns:   font_rm, font_it, etc, or font_unknown if not recognized
 */
 
-void
-font_init(void)
+uint32_t
+font_readtype(BOOL soft)
 {
-font_addbasefont(US"Times-Roman");
-font_addbasefont(US"Times-Italic");
-font_addbasefont(US"Times-Bold");
-font_addbasefont(US"Times-BoldItalic");
-font_addbasefont(US"Symbol");
-font_addbasefont(US"PMW-Music");
+read_nextword();
+if (Ustrcmp(read_wordbuffer, "roman") == 0) return font_rm;
+if (Ustrcmp(read_wordbuffer, "italic") == 0) return font_it;
+if (Ustrcmp(read_wordbuffer, "bold") == 0) return font_bf;
+if (Ustrcmp(read_wordbuffer, "bolditalic") == 0) return font_bi;
+if (Ustrcmp(read_wordbuffer, "symbol") == 0) return font_sy;
+
+if (Ustrcmp(read_wordbuffer, "extra") == 0)
+  {
+  int32_t x;
+  if (!read_expect_integer(&x, FALSE, FALSE)) return font_unknown;
+  if (x >= 1 && x <= MaxExtraFont) return font_xx + x - 1;
+  error(ERR37, MaxExtraFont);
+  }
+else if (!soft)
+  error(ERR8, "\"roman\", \"italic\", \"bold\", \"bolditalic\", \"symbol\", "
+    "or \"extra\"");
+return font_unknown;
 }
 
 
@@ -991,19 +1179,20 @@ font_addbasefont(US"PMW-Music");
 /* This function searches the font list by the name of the font.
 
 Argument:  the font name (e.g. "Times-Roman")
-Returns:   the font id (offset in font_List) or -1 if not found
+Returns:   the font id (offset in font_list) or font_unknown if not found
 */
 
-int
+uint32_t
 font_search(uschar *name)
 {
-int i;
+uint32_t i;
 for (i = 0; i < font_count; i++)
-  if (Ustrcmp(name, font_List[i].psname) == 0) return i;
-return -1;
+  if (Ustrcmp(name, font_list[i].name) == 0) return i;
+return font_unknown;
 }
 
 
+
 /*************************************************
 *      Manually translate Unicode code point     *
 *************************************************/
@@ -1018,10 +1207,10 @@ Arguments:
 Returns:      possibly a different code point
 */
 
-int
-font_utranslate(int c, fontstr *fs)
+uint32_t
+font_utranslate(uint32_t c, fontstr *fs)
 {
-int top, bot, mid;
+int top, bot;
 
 if (fs->utrcount == 0) return c;
 
@@ -1030,7 +1219,7 @@ top = fs->utrcount;
 
 while (bot < top)
   {
-  mid = (bot + top)/2;
+  int mid = (bot + top)/2;
   if (c == fs->utr[mid].unicode) return fs->utr[mid].pscode;
   if (c > fs->utr[mid].unicode) bot = mid + 1;
     else top = mid;
@@ -1042,257 +1231,147 @@ return c;   /* Not found */
 
 
 /*************************************************
-*         Find the width of a string             *
+*         Find the width of a character          *
 *************************************************/
 
-/* The "height" (distance current point moves vertically) is placed in
-font_stringheight. For ordinary text fonts, this is only non-zero when there's
-a rotation, but some characters in the music font move vertically. If
-font_xstretch is set, it is an absolute amount of space to add to each space
-character. This is used when printing justified headings and footings.
+/* The "height" (distance current point moves vertically) is also returned. For
+ordinary text fonts, this is only non-zero when there's a rotation, but some
+characters in the music font move vertically. Such heights apply only to
+low-numbered characters in certain fonts.
+
+All code points in PMW strings are checked for validity at read-in time and
+translated if necessary. By this stage, except for the special non-Unicode
+values, the code points will always be less than 256 for non-standardly-encoded
+fonts, and less than LOWCHARLIMIT plus a few extras for standardly-encoded
+fonts. Either way, the values are less than the size of the widths vector.
+
+This function might be called with a non-Unicode special character if an escape
+such as \p\ or \so\ is used in the first part of a heading. A zero width is
+returned because this is documented not to work when a long heading is being
+split into multiple lines.
 
 Arguments:
-  s            the string
+  c            the character
+  lastc        the previous character or 0 (for kerning)
   font         the font number
-  pointsize    the font's point size
+  pointsize    the font size
+  hptr         where to return the height or NULL
 
 Returns:       the string width
 */
 
-int
-font_stringwidth(uschar *s, int font, int pointsize)
+int32_t
+font_charwidth(uint32_t c, uint32_t lastc, uint32_t font, int32_t pointsize,
+  int32_t *hptr)
 {
-int hwidth = 0;
-int vheight = 0;
-int spacecount = 0;
-int c;
-int lastc = -1;
+fontstr *fs;
+kerntablestr *ktable;
+int32_t yield;
 
-#ifdef SUPPORT_B2PF
-uschar b2pf_stack_buffer[B2PF_STACKBSIZE];
-#endif
+DEBUG(D_font) eprintf("font_charwidth %d %s \'%c\'\n", font, sff(pointsize), c);
 
-int fontid = font_table[font];
-fontstr *fs = &(font_List[fontid]);
-int *widths = fs->widths;
-int *heights = fs->heights;
-tree_node *high_tree = fs->high_tree;
-kerntablestr *ktable = fs->kerns;
-int kerncount = fs->kerncount;
-BOOL stdencoding = fs->stdencoding;
+/* Adjust the font size for small caps and reduced music font. */
 
-DEBUG(("font_stringwidth %d %d \"%s\"\n", fontid, pointsize, s));
-
-/* If we have B2PF support, process this string if its font is set up for it.
-For the moment, just use the on-stack output buffer. If ever there's a
-complaint, we could retry the process using a larger heap buffer. */
-
-#ifdef SUPPORT_B2PF
-if (font_b2pf_contexts[font] != NULL)
+if (font >= font_small)
   {
-  size_t ilen = Ustrlen(s);
-  size_t used, eoffset;
-  int rc;
-
-  /* Reduce output size by 1 to allow for a terminating zero to be added. */
-
-  rc = b2pf_format_string(font_b2pf_contexts[font], (void *)s, ilen,
-    (void *)b2pf_stack_buffer, B2PF_STACKBSIZE - 1, &used,
-    font_b2pf_options[fontid], &eoffset);
-
-  if (rc != B2PF_SUCCESS)
-    {
-    size_t buffused;
-    char buffer[128];
-    (void)b2pf_get_error_message(rc, buffer, sizeof(buffer), &buffused, 0);
-    buffer[buffused] = 0;
-    error_moan(ERR156, s, buffer);  /* Hard */
-    }
-
-  b2pf_stack_buffer[used] = 0;
-  s = b2pf_stack_buffer;
+  font -= font_small;
+  if (font == font_mf) pointsize = (pointsize * 9) / 10;
+    else pointsize = (pointsize * curmovt->smallcapsize) / 1000;
   }
-#endif
 
-/* Now scan the string. For non-standardly encoded fonts there may be a .utr
-translation from Unicode to the font's encoding. When kerning we have to note
-the kern values. */
+/* Get the basic width from the font information. */
 
-while (*s != 0)
+fs = &(font_list[font_table[font]]);
+
+if (c <= MAX_UNICODE)
   {
-  GETCHARINC(c, s);
-  if (c == ' ') spacecount++;
-  if (!stdencoding) c = font_utranslate(c, fs);
+  yield = fs->widths[c];
+  if (hptr != NULL) *hptr = (fs->heights != NULL && c < 256)?
+    mac_muldiv(fs->heights[c], pointsize, 1000) : 0;
+  }
 
-  /* Deal with characters >= 256 */
+else
+  {
+  yield = 0;
+  if (hptr != NULL) *hptr = 0;
+  }
 
-  if (c >= 256)
+/* Deal with kerning. At the moment, this is supported only for characters
+whose code points are less than 0xffff. It should be easy to extend by
+converting to 64-bit combinations. */
+
+ktable = fs->kerns;
+if (main_kerning && ktable != NULL &&
+    lastc > 0 && lastc <= 0xffff && c <= 0xffff)
+  {
+  uint32_t pair = (lastc << 16) | c;
+  int top = fs->kerncount;
+  int bot = 0;
+
+  while (top > bot)
     {
-    if (!stdencoding)
-      hwidth += widths[UNKNOWN_CHAR_N];
-
-    else if (c < LOWCHARLIMIT)
-      hwidth += widths[c];
-
-    else
+    int mid = (top + bot)/2;
+    kerntablestr *k = &(ktable[mid]);
+    if (pair == k->pair)
       {
-      tree_node *t;
-      uschar utf[8];
-      utf[misc_ord2utf8(c, utf)] = 0;
-      t = Tree_Search(high_tree, utf);
-      if (t != NULL) hwidth += t->val[1];
-        else hwidth += widths[UNKNOWN_CHAR_S];
+      yield += k->kwidth;
+      break;
       }
+    if (pair > k->pair) bot = mid + 1; else top = mid;
     }
-
-  /* Character is < 256. Heights apply only to low-numbered characters. */
-
-  else
-    {
-    hwidth += widths[c];
-    if (heights != NULL) vheight += heights[c];
-    }
-
-  /* Deal with kerning */
-
-  if (main_kerning && ktable != NULL &&
-      lastc >= 0 && lastc <= 0xffff && c <= 0xffff)
-    {
-    int top, bot, mid;
-    usint pair = (usint)((usint)lastc << 16) | c;
-
-    bot = 0;
-    top = kerncount;
-
-    while (top > bot)
-      {
-      kerntablestr *k;
-      mid = (top + bot)/2;
-      k = &(ktable[mid]);
-      if (pair == k->pair)
-        {
-        hwidth += k->kwidth;
-        break;
-        }
-      if (pair > k->pair) bot = mid + 1; else top = mid;
-      }
-    }
-
-  lastc = c;
   }
 
-/* Scale to font size, then adjust for rotation */
-
-hwidth = mac_muldiv(hwidth, pointsize, 1000);
-vheight = mac_muldiv(vheight, pointsize, 1000);
-
-font_stringheight =
-  mac_muldiv(hwidth, font_transform[1], 65536) +
-  mac_muldiv(vheight, font_transform[3], 65536);
-
-DEBUG(("font_stringwidth() %d %d\n",
-       mac_muldiv(hwidth, font_transform[0], 65536) +
-       mac_muldiv(vheight, font_transform[2], 65536),
-       font_stringheight));
-
-return font_xstretch * spacecount +
-       mac_muldiv(hwidth, font_transform[0], 65536) +
-       mac_muldiv(vheight, font_transform[2], 65536);
-}
-
-
-
-
-/*************************************************
-*            Read a font word (+ number)         *
-*************************************************/
-
-/*
-Argument:  TRUE to not give an error if the word is unrecognised
-Returns:   font_rm, font_it, etc, or -1 on error
-*/
-
-int
-font_fontword(BOOL soft)
-{
-next_word();
-if (Ustrcmp(read_word, "roman") == 0) return font_rm;
-if (Ustrcmp(read_word, "italic") == 0) return font_it;
-if (Ustrcmp(read_word, "bold") == 0) return font_bf;
-if (Ustrcmp(read_word, "bolditalic") == 0) return font_bi;
-if (Ustrcmp(read_word, "symbol") == 0) return font_sy;
-if (Ustrcmp(read_word, "music") == 0) return font_mu;
-if (Ustrcmp(read_word, "bigmusic") == 0) return font_mf;
-if (Ustrcmp(read_word, "extra") == 0)
-  {
-  int x;
-  if (!read_expect_integer(&x, FALSE, FALSE)) return -1;
-  if (x >= 1 && x <= MaxExtraFont) return font_xx + x - 1;
-  error_moan(ERR26, MaxExtraFont);
-  }
-else if (!soft)
-  error_moan(ERR10, "\"roman\", \"italic\", \"bold\", \"bolditalic\", or \"extra\"");
-return -1;
+return mac_muldiv(yield, pointsize, 1000);
 }
 
 
 
 /*************************************************
-*         Set up a font rotation                 *
+*              Rotate a font matrix              *
 *************************************************/
 
-/* The effect of this function is to modify the font_transform matrix by the
-given rotation.
+/* This function is called when a text string is to be rotated when output. It
+returns a pointer to a static copy of the input font instance block, with a
+suitably rotated matrix setting.
 
-Argument:  rotation in millidegrees
-Returns:   nothing
+Arguments:
+  fdata       pointer to a font instance block
+  angle       rotation angle in degrees
+
+Returns:      pointer to a copy of fdata with rotation set
 */
 
-void
-font_rotate(int rotate)
-{
-int newmatrix[4];
-double sr = sin(((double)rotate)*atan(1.0)/45000.0);
-double cr = cos(((double)rotate)*atan(1.0)/45000.0);
+static fontinststr copyfdata;
+static int32_t copymatrix[6];
+static int32_t nullmatrix[6] = { 65536, 0, 0, 65536, 0, 1000 };
 
-font_sinr = (int)(sr * 1000.0);
-font_cosr = (int)(cr * 1000.0);
+fontinststr *
+font_rotate(fontinststr *fdata, int32_t angle)
+{
+int32_t newmatrix[4];
+double sr = sin(((double)angle)*atan(1.0)/45000.0);
+double cr = cos(((double)angle)*atan(1.0)/45000.0);
+
+copyfdata = *fdata;
+copyfdata.matrix = copymatrix;
+
+memcpy(copymatrix, (fdata->matrix == NULL)? nullmatrix : fdata->matrix,
+  4 * sizeof(int32_t));
 
 newmatrix[0] =
-  (int)(((double)font_transform[0])*cr - ((double)font_transform[1])*sr);
+  (int)(((double)copymatrix[0])*cr - ((double)copymatrix[1])*sr);
 newmatrix[1] =
-  (int)(((double)font_transform[0])*sr + ((double)font_transform[1])*cr);
+  (int)(((double)copymatrix[0])*sr + ((double)copymatrix[1])*cr);
 newmatrix[2] =
-  (int)(((double)font_transform[2])*cr - ((double)font_transform[3])*sr);
+  (int)(((double)copymatrix[2])*cr - ((double)copymatrix[3])*sr);
 newmatrix[3] =
-  (int)(((double)font_transform[2])*sr + ((double)font_transform[3])*cr);
+  (int)(((double)copymatrix[2])*sr + ((double)copymatrix[3])*cr);
 
-memcpy(font_transform, newmatrix, 4*sizeof(int));
+memcpy(copymatrix, newmatrix, 4*sizeof(int32_t));
+copymatrix[4] = (int32_t)(sr * 1000.0);
+copymatrix[5] = (int32_t)(cr * 1000.0);
+return &copyfdata;
 }
 
 
-
-/*************************************************
-*           Reset font transformation            *
-*************************************************/
-
-/* Historical relic: the transform matrix operates in the units that RISC OS
-uses for its font transformations. These were chosen so that the RISC OS screen
-display code could use the matrix directly. There seems no reason to change
-this choice gratuitously. The first four matrix values have 16-bit fractions;
-the other two are in millipoints.
-
-Arguments:  none
-Returns:    nothing
-*/
-
-void font_reset(void)
-{
-font_sinr = 0;      /* No rotation => angle is zero */
-font_cosr = 1000;   /* Fixed point 1.0 */
-font_transform[0] = font_transform[3] = 65536;
-font_transform[1] = font_transform[2] = 0;
-font_transform[4] = font_transform[5] = 0;
-}
-
 /* End of font.c */
