@@ -1,32 +1,21 @@
 /*************************************************
-*    The PMW Music Typesetter - 3rd incarnation  *
+*              PMW beaming functions             *
 *************************************************/
 
-/* Copyright (c) Philip Hazel, 1991 - 2020 */
-
-/* Written by Philip Hazel, starting November 1991 */
-/* This file last modified: April 2020 */
-
-
-/* This file contains code for setting up and outputting beams */
-
-
-#include "pmwhdr.h"
-#include "pagehdr.h"
-#include "readhdr.h"
-#include "outhdr.h"
+/* Copyright Philip Hazel 2021 */
+/* This file created: June 2021 */
+/* This file last modified: July 2021 */
 
+#include "pmw.h"
 
 
 /*************************************************
 *              Static variables                  *
 *************************************************/
 
+static int32_t retryslopes[] = { 100, 50, 0, -50, -100, -300 };
 
-static int retryslopes[] = { 100, 50, 0, -50, -100, -300 };
-
-static int minoffset;
-
+static int32_t minoffset;
 
 
 
@@ -42,18 +31,34 @@ Arguments:   none
 Returns:     slope
 */
 
-static int
+static int32_t
 findslope(void)
 {
 b_notestr *left = beam_first;
+
+int32_t slope = 0;
+
+/* Pitches in notes are uint16_t, but here we are going to be working with
+positive and negative intervals, so use ints. */
+
 int leftpitch = left->spitch;
 int rightpitch = beam_last->spitch;
-int slope, thisrun, runindex, lastrunindex, lasttype;
+int thispitch = leftpitch;
+int nextpitch;
+int max = 0;
+int min = INT_MAX;
+
+int thisrun = 0;
+int runindex;
+int lastrunindex = -1;
+int lasttype = left->notetype;
 int interval;
 int lastinterval = 0;
-int thispitch, nextpitch, max, min, firstslope;
+int firstslope = beam_upflag? (-1) : (+1);   /* in case rest first */
 int runs[3];
-BOOL downs_less_than_up, ups_less_than_down;
+
+BOOL downs_less_than_up = TRUE;
+BOOL ups_less_than_down = TRUE;
 
 mac_couplepitch(leftpitch, left->flags);
 mac_couplepitch(rightpitch, beam_last->flags);
@@ -64,19 +69,11 @@ if (leftpitch == rightpitch) return 0;
 
 /* Otherwise do some analysis */
 
-firstslope = beam_upflag? (-1) : (+1);   /* in case rest first */
-slope = thisrun = 0;
-thispitch = leftpitch;
-lastrunindex = -1;
-lasttype = left->notetype;
-max = 0;
-min = BIGNUMBER;
 runs[0] = runs[1] = runs[2] = 0;                 /* equal, up, down */
-downs_less_than_up = ups_less_than_down = TRUE;
 
 while (left != beam_last)
   {
-  b_notestr *next = misc_nextnote(left, NULL);
+  b_notestr *next = misc_nextnote(left);
   nextpitch = next->spitch;
 
   /* Totally ignore grace/non-grace as required */
@@ -88,7 +85,7 @@ while (left != beam_last)
     continue;
     }
 
-  /* Skip over rests, but keep the note type */
+  /* Skip over rests, but keep the note type (below) */
 
   if (nextpitch != 0)
     {
@@ -143,22 +140,23 @@ if (lastrunindex >= 0 && thisrun > runs[lastrunindex])
 
 /* This is the heuristic algorithm that produces a default slope */
 
-if (runs[1] > runs[0] && runs[1] > runs[2]) slope = curmovt->maxbeamslope2;
-  else if (runs[2] > runs[0] && runs[2] > runs[1]) slope = - curmovt->maxbeamslope2;
+if (runs[1] > runs[0] && runs[1] > runs[2]) slope = curmovt->maxbeamslope[1];
+  else if (runs[2] > runs[0] && runs[2] > runs[1])
+    slope = -curmovt->maxbeamslope[1];
 
 /* Some special cases override the heuristic */
 
 if (leftpitch > rightpitch)
   {
-  if (beam_upflag && leftpitch - rightpitch >= 4 && max <= rightpitch + 2)
-    slope = - curmovt->maxbeamslope2;
+  if (beam_upflag && leftpitch - rightpitch >= 8 && max <= rightpitch + 4)
+    slope = -curmovt->maxbeamslope[1];
   else if (!beam_upflag && min <= rightpitch) slope = 0;
   else if (slope > 0) slope = 0;
   }
 else
   {
-  if (!beam_upflag && rightpitch - leftpitch >= 4 && min >= rightpitch - 2)
-    slope = curmovt->maxbeamslope2;
+  if (!beam_upflag && rightpitch - leftpitch >= 8 && min >= rightpitch - 4)
+    slope = curmovt->maxbeamslope[1];
   else if (beam_upflag && max >= rightpitch) slope = 0;
   else if (slope < 0) slope = 0;
   }
@@ -190,7 +188,6 @@ sets of notes. If this is successful, then a middle position is chosen at the
 end.
 
 Arguments:
-  ornament       the first note's ornament, or -1
   lastopposite   TRUE if the last note's stem goes the opposite way
   adjusts        x-adjustments for the notes
   longestnote    longest note in the beam
@@ -200,25 +197,43 @@ Returns:         FALSE if notes on opposite sides are wanted, but cannot be
 */
 
 static BOOL
-ComputeBeamOffset(int ornament, BOOL lastopposite, int *adjusts,
-  int longestnote)
+ComputeBeamOffset(BOOL lastopposite, int32_t *adjusts, int longestnote)
 {
 b_notestr *thisnote = beam_first;
 BOOL forcelength = TRUE;
-int diff_thisoffset;
-int diff_thatoffset = BIGNUMBER;            /* => unset */
-int hemidemisemiquavers = FALSE;
-int moff = beam_firstmoff;
+BOOL hemidemisemiquavers = FALSE;
+int32_t diff_thisoffset;
+int32_t diff_thatoffset = INT32_MAX;            /* => unset */
+int32_t moff = beam_firstmoff;
+int32_t xfirst = out_findXoffset(beam_firstmoff) + adjusts[0];
+int32_t yadjust = beam_first->yextra;
+int32_t yfirst;
 int adjustptr = 0;
-int xfirst = out_findXoffset(beam_firstmoff) + adjusts[0];
-int yadjust = beam_first->yextra;
 int yfirstpitch = beam_first->spitch;
-int yfirst;
+int tremolo = -1;
+
+/* Note that yadjust does not include a value from "stemlengths", since those
+values are mainly aimed at non-beamed notes. Impose a minimum shortening, just
+in case someone specifies a ridiculous stem length with the note. */
+
+if (yadjust < -8000) yadjust = -8000;
 
 mac_couplepitch(yfirstpitch, beam_first->flags);
-yfirst = (yfirstpitch - 130)*main_stavemagn;
+yfirst = (yfirstpitch - P_1S)*out_pitchmagn;
 beam_splitOK = TRUE;
 
+/* Scan any ornaments for tremolos */
+
+if (n_ornament != NULL)
+  {
+  for (b_ornamentstr *op = n_ornament; op->type == b_ornament;
+       op = (b_ornamentstr *)(op->next))
+    {
+    if (op->ornament == or_trem2 || op->ornament == or_trem3)
+      tremolo = op->ornament;
+    }
+  }
+
 /* Initialize from the first note, forcing a stem length if not coupled. */
 
 if (beam_upflag)
@@ -227,8 +242,8 @@ if (beam_upflag)
   if ((beam_first->flags & nf_coupleD) == 0)
     {
     forcelength = FALSE;
-    if (yfirstpitch < 122)
-      beam_offset += (122 - yfirstpitch)*main_stavemagn;
+    if (yfirstpitch < P_0L - P_T)  /* Space below 0th line */
+      beam_offset += (P_0L - P_T - yfirstpitch)*out_pitchmagn;
     }
   }
 else
@@ -237,8 +252,8 @@ else
   if ((beam_first->flags & nf_coupleU) == 0)
     {
     forcelength = FALSE;
-    if (yfirstpitch > 150)
-      beam_offset -= (yfirstpitch - 150)*main_stavemagn;
+    if (yfirstpitch > P_6S)
+      beam_offset -= (yfirstpitch - P_6S)*out_pitchmagn;
     }
   }
 
@@ -249,35 +264,39 @@ minoffset = abs(beam_offset) + 8000;
 for (;;)
   {
   BOOL note_upflag = (thisnote->flags & nf_stemup) != 0;
-  int xnote = out_findXoffset(moff) -
+  int32_t xnote = out_findXoffset(moff) -
     ((lastopposite && thisnote == beam_last)? beam_Xcorrection : 0) +
       adjusts[adjustptr];
-  int thatynote;
-  int ybasic = yfirst + mac_muldiv(beam_slope, xnote-xfirst, 1000);
-  int ybeam = ybasic + beam_offset;
-  int thisybeam = ybasic + diff_thisoffset;
+  int32_t thatynote;
+  int32_t ybasic = yfirst + mac_muldiv(beam_slope, xnote-xfirst, 1000);
+  int32_t ybeam = ybasic + beam_offset;
+  int32_t thisybeam = ybasic + diff_thisoffset;
+  int32_t ynote;
   int ypitch = thisnote->spitch;
-  int ynote;
 
   mac_couplepitch(ypitch, thisnote->flags);
 
   yadjust = thisnote->yextra;
+  if (yadjust < -8000) yadjust = -8000;
 
-  if (ornament == or_trem2)
-    yadjust += ((ypitch & 2) == 0)? 4000 : 2000;
+  /* Adjust for tremolos: if the note is in a space, the pitch will have the 4
+  bit set. */
 
-  else if (ornament == or_trem3)
-    yadjust += ((ypitch & 2) == 0)? 8000 : 6000;
+  if (tremolo == or_trem2)
+    yadjust += ((ypitch & 4) == 0)? 4000 : 2000;
 
-  ynote = (ypitch - 130)*main_stavemagn + n_upfactor*yadjust;
+  else if (tremolo == or_trem3)
+    yadjust += ((ypitch & 4) == 0)? 8000 : 6000;
+
+  ynote = (ypitch - P_1S)*out_pitchmagn + n_upfactor*yadjust;
 
   /* If this is the first note of the beam, it might really be a rest in
-  disguise. We can tell this from the true pitch. If so, adjust for
+  disguise. We can tell this from the absolute pitch. If so, adjust for
   longestnote. */
 
-  if (thisnote->truepitch == 0)
+  if (thisnote->abspitch == 0)
     {
-    ynote = ynote + n_upfactor*(longestnote - quaver)*4*main_stavemagn;
+    ynote = ynote + n_upfactor*(longestnote - quaver)*4*out_stavemagn;
     }
 
   /* Else if this is the first non-coupled note, force a stem length. We don't
@@ -299,7 +318,7 @@ for (;;)
       if ((thisnote->flags & nf_coupleU) == 0)
         {
         forcelength = FALSE;
-        if (ynote > L_6L + 2000) ynote -= ynote - (L_6L+2000);
+        if (ynote > L_6L + 2000) ynote -= ynote - (L_6L + 2000);
         }
       }
     }
@@ -310,14 +329,14 @@ for (;;)
 
   if (beam_upflag)
     {
-    int m;
+    int32_t m;
     if (ynote > ybeam) beam_offset += ynote - ybeam;
     m = ybasic + beam_offset - ynote + yadjust + 8000;
     if (m < minoffset) minoffset = m;
     }
   else
     {
-    int m;
+    int32_t m;
     if (ynote < ybeam) beam_offset -= ybeam - ynote;
     m = ynote - (ybasic + beam_offset) + yadjust + 8000;
     if (m < minoffset) minoffset = m;
@@ -334,12 +353,12 @@ for (;;)
 
   else if (beam_splitOK)
     {
-    int thatybeam;
+    int32_t thatybeam;
 
     /* If this is the first opposed note, calculate an initial offset, else
     adjust the offset. */
 
-    if (diff_thatoffset == BIGNUMBER)
+    if (diff_thatoffset == INT32_MAX)
       diff_thatoffset = (thatynote - n_upfactor*28000) - ybasic;
     else
       {
@@ -355,7 +374,7 @@ for (;;)
 
   /* If this is a split beam, check that it's still possible */
 
-  if (beam_splitOK && diff_thatoffset != BIGNUMBER &&
+  if (beam_splitOK && diff_thatoffset != INT32_MAX &&
     ((beam_upflag && diff_thatoffset < diff_thisoffset) ||
       (!beam_upflag && diff_thatoffset > diff_thisoffset)))
         return FALSE;
@@ -372,10 +391,21 @@ for (;;)
     {
     adjustptr++;
     moff += thisnote->length;
-    thisnote = misc_nextnote(thisnote, &ornament);
 
-    if (out_gracenotes) moff++;
-      else if (thisnote->length == 0) continue;
+    thisnote = misc_nextnote(thisnote);
+    tremolo = -1;
+
+    if (main_nextnoteornament != NULL)
+      {
+      for (b_ornamentstr *op = main_nextnoteornament; op->type == b_ornament;
+           op = (b_ornamentstr *)(op->next))
+        {
+        if (op->ornament == or_trem2 || op->ornament == or_trem3)
+          tremolo = op->ornament;
+        }
+      }
+
+    if (out_gracenotes) moff++; else if (thisnote->length == 0) continue;
 
     /* Not a rest - go round the outer loop again */
 
@@ -389,18 +419,16 @@ for (;;)
 
     if (beam_upflag)
       {
-      int yrest = (thisnote->yextra*main_stavemagn)/1000;
-      if (thisnote->notetype > squaver) yrest += 4*main_stavemagn;
-      if (longestnote > quaver)
-        yrest += (longestnote - quaver)*4*main_stavemagn;
+      int32_t yrest = (thisnote->yextra*out_stavemagn)/1000;
+      if (thisnote->notetype > squaver) yrest += 4*out_stavemagn;
+      if (longestnote > quaver) yrest += (longestnote - quaver)*4*out_stavemagn;
       if (yrest > ybeam) beam_offset += yrest - ybeam;
       }
     else
       {
-      int yrest = 12*main_stavemagn + (thisnote->yextra*main_stavemagn)/1000;
-      if (thisnote->notetype > quaver) yrest -= 4*main_stavemagn;
-      if (longestnote > quaver)
-        yrest -= (longestnote - quaver)*4*main_stavemagn;
+      int32_t yrest = 12*out_stavemagn + (thisnote->yextra*out_stavemagn)/1000;
+      if (thisnote->notetype > quaver) yrest -= 4*out_stavemagn;
+      if (longestnote > quaver) yrest -= (longestnote - quaver)*4*out_stavemagn;
       if (yrest < ybeam) beam_offset -= ybeam - yrest;
       }
 
@@ -419,15 +447,14 @@ if (hemidemisemiquavers)
   {
   beam_offset += n_upfactor*1000;
   diff_thisoffset += n_upfactor*1000;
-  if (diff_thatoffset != BIGNUMBER) diff_thatoffset -= n_upfactor*1000;
+  if (diff_thatoffset != INT32_MAX) diff_thatoffset -= n_upfactor*1000;
   }
 
 /* If notes on the other side of the beam have been successfully processed, use
 the compromise offset. The flag is still TRUE if none have been encountered, so
-we set it FALSE to save wasting time with additional processing in later
-routines. */
+we set it FALSE to save wasting time with additional processing later. */
 
-if (beam_splitOK && diff_thatoffset != BIGNUMBER)
+if (beam_splitOK && diff_thatoffset != INT32_MAX)
   beam_offset = (diff_thisoffset + diff_thatoffset)/2;
     else beam_splitOK = FALSE;    /* this is the normal case! */
 
@@ -454,17 +481,16 @@ Returns:         nothing
 */
 
 static void
-DrawBeams(int xright, int *adjusts, int longestnote, BOOL lastopposite,
+DrawBeams(int32_t xright, int32_t *adjusts, int longestnote, BOOL lastopposite,
   BOOL rightbreak2)
 {
-int level, i;
-int lastlen = (len_minim*out_pnum)/out_pden;
+int lastlen = (len_minim*out_pletnum)/out_pletden;
 int throughlevel = longestnote - quaver + 1;
 int thisnotetype = longestnote + 1;
-int sladjust = mac_muldiv(75*main_stavemagn, n_fontsize, 1000000) + n_cueadjust;
-int sradjust = mac_muldiv(115*main_stavemagn, n_fontsize, 1000000) + n_cueadjust;
+int32_t sladjust = mac_muldiv(75*out_stavemagn, n_fontsize, 1000000) + n_cueadjust;
+int32_t sradjust = mac_muldiv(115*out_stavemagn, n_fontsize, 1000000) + n_cueadjust;
 
-uschar taildirflags[MAX_BEAMSIZE];
+uint8_t taildirflags[MAX_BEAMNOTES];
 
 /* Add possible cue note adjustment into endpoints */
 
@@ -475,7 +501,7 @@ xright += n_cueadjust;
 
 if (n_fontsize != 10000)
   {
-  int adjust = (14*(10000 - n_fontsize))/10;
+  int32_t adjust = (14*(10000 - n_fontsize))/10;
   if (beam_upflag) beam_firstY -= adjust;
     else beam_firstY += adjust;
   }
@@ -486,17 +512,17 @@ beam, as it then correctly sets the values to zero. */
 
 if (beam_splitOK)
   {
-  int adjust = (throughlevel-1)*3;
+  int adjust = (throughlevel - 1)*3;
   b_notestr *p = beam_first;
 
-  for (i = beam_count; i >= 1 ; i--)
+  for (int i = beam_count; i >= 1 ; i--)
     {
     beam_stemadjusts[i] =
       (((p->flags & nf_stemup) != 0) == beam_upflag)? 0 : adjust;
-    p = misc_nextnote(p, NULL);
+    p = misc_nextnote(p);
     while (p != NULL && p->spitch == 0)
       {
-      p = misc_nextnote(p, NULL);
+      p = misc_nextnote(p);
       i--;
       }
     }
@@ -511,7 +537,7 @@ if (beam_accrit != 0)
   int count = abs(beam_accrit);
   int sign = beam_accrit/count;
 
-  for (i = 0; i < count; i++)
+  for (int i = 0; i < count; i++)
     {
     ps_beam(beam_firstX, xright, 1, i*sign);
     }
@@ -521,7 +547,7 @@ if (beam_accrit != 0)
 
 /* "Normal" beams - first draw beams which go right through all notes */
 
-for (i = 1; i <= throughlevel; i++)
+for (int i = 1; i <= throughlevel; i++)
   {
   ps_beam(beam_firstX, xright, i, 0);
   lastlen /= 2;
@@ -529,50 +555,48 @@ for (i = 1; i <= throughlevel; i++)
 
 /* Clear the flags which remember about subtail directions and opposites */
 
-for (i = 0; i <= beam_count; i++) taildirflags[i] = 0;
+for (int i = 0; i <= beam_count; i++) taildirflags[i] = 0;
 
 /* Now scan the notes and add internal beams or beamlets - thisnotetype
 contains the notetype for the beam under consideration. */
 
-for (level = throughlevel+1; level <= 4; level++)
+for (int level = throughlevel + 1; level <= 4; level++)
   {
   b_notestr *left = beam_first;
   b_notestr *postleft = left;
 
-  int donesomething = FALSE;
-  int notenumber = 1;
-  int leftmoff = beam_firstmoff;
-  int rightmoff = leftmoff;
   int adjustptr = 0;
-  int firstbreak2 = FALSE;
-  int leftdone = FALSE;
-  int leftextended = FALSE;
-  int rightextended = FALSE;
+  int notenumber = 1;
+  int32_t leftmoff = beam_firstmoff;
+  int32_t rightmoff = leftmoff;
+
+  BOOL donesomething = FALSE;
+  BOOL firstbreak2 = FALSE;
+  BOOL leftdone = FALSE;
+  BOOL leftextended = FALSE;
+  BOOL rightextended = FALSE;
 
   /* Determine whether there is a second-level break after the very first note
-  - this is a special (unusual) case */
+  - this is a special (unusual) case. */
 
-  do mac_advancechord(postleft); while (postleft->type == b_chord);
-  if (postleft->type == b_tie)
-    {
-    postleft = (b_notestr *)((uschar *)postleft + length_table[b_tie]);
-    if (postleft->type == b_Jump)
-      postleft = (b_notestr *)
-        ((uschar *)(((b_Jumpstr *)postleft)->next) + length_table[b_Jump]);
-    }
-  firstbreak2 = (postleft->type == b_beambreak2);
+  do postleft = (b_notestr *)(postleft->next);
+    while (postleft->type == b_chord || postleft->type == b_tie);
+  firstbreak2 = postleft->type == b_beambreak &&
+    ((b_beambreakstr *)postleft)->value != BEAMBREAK_ALL;
 
   /* Now scan the beam */
 
   while (left != beam_last)
     {
-    b_notestr *right = misc_nextnote(left, NULL);
+    b_notestr *right = misc_nextnote(left);
     b_notestr *postright;
 
     int rightadjustptr = adjustptr + 1;
-    int left_up = (left->flags & nf_stemup) != 0;
     int leftnotetype = left->notetype;
-    int rightnotetype, right_up;
+    int rightnotetype;
+
+    BOOL left_up = (left->flags & nf_stemup) != 0;
+    BOOL right_up;
 
     rightmoff += left->length;
 
@@ -585,7 +609,7 @@ for (level = throughlevel+1; level <= 4; level++)
            (right->length == 0 && !out_gracenotes))
       {
       rightmoff += right->length;
-      right = misc_nextnote(right, NULL);
+      right = misc_nextnote(right);
       rightadjustptr++;
       }
 
@@ -596,32 +620,29 @@ for (level = throughlevel+1; level <= 4; level++)
     rightnotetype = right->notetype;
 
     postright = right;
-    do mac_advancechord(postright); while (postright->type == b_chord);
-    if (postright->type == b_tie)
-      {
-      postright = (b_notestr *)((uschar *)postright + length_table[b_tie]);
-      if (postright->type == b_Jump)
-        postright = (b_notestr *)
-          ((uschar *)(((b_Jumpstr *)postright)->next) + length_table[b_Jump]);
-      }
-    rightbreak2 = (postright->type == b_beambreak2);
+    do postright = (b_notestr *)(postright->next);
+      while (postright->type == b_chord || postright->type == b_tie);
+
+    rightbreak2 = postright->type == b_beambreak &&
+      ((b_beambreakstr *)postright)->value != BEAMBREAK_ALL;
 
     /* Check for rest beyond right note and see if it requires a beam break. */
 
     if (right != beam_last)
       {
-      b_notestr *afterright = misc_nextnote(right, NULL);
+      b_notestr *afterright = misc_nextnote(right);
       if (afterright->spitch == 0 &&
         afterright->notetype < rightnotetype) rightbreak2 = TRUE;
       }
 
-    /* If both notes longer than current, do nothing */
+    /* Something gets drawn only if at least one of the notes is shorter or
+    equal to the current note type. */
 
     if (leftnotetype >= thisnotetype || rightnotetype >= thisnotetype)
       {
-      int skipnote = FALSE;
-      int XLadjust = 0;
-      int XRadjust = 0;
+      BOOL skipnote = FALSE;
+      int32_t XLadjust = 0;
+      int32_t XRadjust = 0;
       int Dlevel = level;
 
       donesomething = TRUE;    /* note to keep going */
@@ -639,9 +660,9 @@ for (level = throughlevel+1; level <= 4; level++)
           (left_up == right_up || leftnotetype == rightnotetype) &&
           (right->spitch != 0 || !beam_overbeam || !out_lineendflag))
         {
+        BOOL useopposite = FALSE;
         int ns = beam_count - notenumber;   /* for addressing stemadjusts */
-        int useopposite = FALSE;
-        int xr;
+        int32_t xr;
 
         /* If the first note already has a beamlet drawn on the opposite side,
         then we must use the opposite side. */
@@ -697,10 +718,10 @@ for (level = throughlevel+1; level <= 4; level++)
           }
 
         /* Make adjustments to the x positions in all cases. Note that we test
-        for a left-hand rest by the true pitch, as its spitch has been fudged
-        to be like a real note. */
+        for a left-hand rest by the absolute pitch, as its spitch has been
+        fudged to be like a real note. */
 
-        if (left_up && left->truepitch != 0) XLadjust = beam_Xcorrection;
+        if (left_up && left->abspitch != 0) XLadjust = beam_Xcorrection;
         if (right_up || right->spitch == 0) XRadjust = beam_Xcorrection;
 
         /* If the right note is a rest, we must be at the end of a beam,
@@ -730,17 +751,15 @@ for (level = throughlevel+1; level <= 4; level++)
         {
         /* If the left-hand note requires a beam at this level, then give it
         the part beam, pointing right, unless its beam has already been drawn.
-        Note that the rest test must be on truepitch. We must set the correct
-        flag for the final type, because on the screen the ends get moved
-        left/right as a result of it (and printer driver output). */
+        Note that the rest test must be on abspitch. */
 
-        if (leftnotetype >= thisnotetype && !leftdone && left->truepitch != 0)
+        if (leftnotetype >= thisnotetype && !leftdone && left->abspitch != 0)
           {
-          int x = out_findXoffset(leftmoff) + adjusts[adjustptr] +
+          int32_t x = out_findXoffset(leftmoff) + adjusts[adjustptr] +
             sladjust + (left_up? beam_Xcorrection : 0);
 
           if (left_up != beam_upflag) Dlevel = throughlevel - level + 1;
-          ps_beam(x, x + mac_muldiv(curmovt->beamflaglength, main_stavemagn,
+          ps_beam(x, x + mac_muldiv(curmovt->beamflaglength, out_stavemagn,
             1000), Dlevel, 0);
           taildirflags[notenumber] |= 1;
           }
@@ -751,15 +770,15 @@ for (level = throughlevel+1; level <= 4; level++)
 
         if (rightnotetype >= thisnotetype && right->spitch != 0 &&
           (right == beam_last || rightbreak2 ||
-            ((misc_nextnote(right, NULL))->notetype < rightnotetype &&
+            ((misc_nextnote(right))->notetype < rightnotetype &&
               (rightmoff%lastlen != 0 ||
                 (taildirflags[notenumber+1] & 3) == 2))))
           {
-          int x = out_findXoffset(rightmoff) + adjusts[rightadjustptr] +
+          int32_t x = out_findXoffset(rightmoff) + adjusts[rightadjustptr] +
             sradjust + (right_up? beam_Xcorrection : 0);
 
           if (right_up != beam_upflag) Dlevel = throughlevel - level + 1;
-          ps_beam(x - mac_muldiv(curmovt->beamflaglength, main_stavemagn, 1000),
+          ps_beam(x - mac_muldiv(curmovt->beamflaglength, out_stavemagn, 1000),
             x, Dlevel, 0);
           taildirflags[notenumber+1] |= 2;
           skipnote = TRUE;
@@ -774,14 +793,14 @@ for (level = throughlevel+1; level <= 4; level++)
         {
         left = right;
         leftmoff = rightmoff;
-        right = misc_nextnote(right, NULL);
+        right = misc_nextnote(right);
         rightmoff += left->length;
         rightadjustptr++;
         while ((right->spitch == 0 && right != beam_last) ||
                (right->length == 0 && !out_gracenotes))
           {
           rightmoff += right->length;
-          right = misc_nextnote(right, NULL);
+          right = misc_nextnote(right);
           rightadjustptr++;
           }
         leftdone = FALSE;
@@ -822,9 +841,9 @@ for (level = throughlevel+1; level <= 4; level++)
 
   if (rightbreak2 && left->notetype >= thisnotetype && left->spitch != 0)
     {
-    int x = out_findXoffset(leftmoff) + adjusts[adjustptr] + sradjust +
+    int32_t x = out_findXoffset(leftmoff) + adjusts[adjustptr] + sradjust +
       ((beam_upflag && !lastopposite)? beam_Xcorrection: 0);
-    ps_beam(x - mac_muldiv(curmovt->beamflaglength, main_stavemagn, 1000), x,
+    ps_beam(x - mac_muldiv(curmovt->beamflaglength, out_stavemagn, 1000), x,
       level, 0);
     }
 
@@ -864,74 +883,67 @@ Returns:        TRUE if beaming is set up; FALSE otherwise.
 */
 
 BOOL
-out_setupbeam(b_notestr *p, int moff, BOOL nodraw, BOOL restatstart)
+out_setupbeam(b_notestr *p, int32_t moff, BOOL nodraw, BOOL restatstart)
 {
-b_notestr *pp = p + 1;
+b_notestr *pp;
+
+int adjustptr = 1;
+int longestnote = p->notetype;    /* A secondary beam break forces this */
+int overbarcount = 0;
+int pendingrests = 0;
+int trailingrests = 0;
+
+int32_t adjusts[MAX_BEAMNOTES];
+int32_t currentmoff = moff;
+int32_t lastadjust = 0;
+int32_t lastmoff = 0;
+int32_t lastmoffinbar = 0;
+int32_t move_adjust = 0;
+int32_t xright, yright;
+
 b_notestr *lastp = p;
 b_notestr *lastinbar = NULL;
 
-int overbarcount = 0;
-int type = pp->type;
-int currentmoff = moff;
-
-int longestnote = p->notetype;    /* a secondary beam break forces this */
-int i, xright, yright;
-
 BOOL lastopposite;
 BOOL rightbreak2 = FALSE;
 
-int adjust = 0;
-int lastadjust = 0;
-int lastmoff = 0;
-int lastmoffinbar = 0;
-int pendingrests = 0;
-int trailingrests = 0;
-int adjustptr = 1;
-int adjusts[50];
-
-DEBUG(("out_setupbeam() start\n"));
-
-adjusts[0] = out_Xadjustment;
+adjusts[0] = out_Xadjustment;   /* From preceding [move] */
 
 beam_count = 1;
 beam_first = beam_last = p;
 beam_firstmoff = beam_lastmoff = moff;
 beam_overbeam = FALSE;
-
-/* The stem direction is determined when the notes are read in. */
-
 beam_upflag = n_upflag;
 
+TRACE("out_setupbeam() start\n");
+
 /* Scan to find end of beam. We have to use gotos to jump out of the loop from
 within the switch. */
 
-for (;;)
+for (pp = (b_notestr *)(p->next); pp->type != b_barline;
+     pp = (b_notestr *)(pp->next))
   {
-  int included;
+  BOOL included;
+  b_beambreakstr *bb;
 
-  switch (type)
+  switch (pp->type)
     {
-    case b_End:
-    if (!((b_Endstr *)pp)->overbeam || beam_overbeam) goto ENDBEAM;
+    case b_barline:
+    goto ENDBEAM;
 
-    /* The bar line has been flagged for continuation of a beam over it. If a
-    subsequent bar exists, its data will follow on immediately after this
-    one's, so we can just allow the continuation to happen, after noting the
-    fact. */
+    case b_overbeam:
+    if (beam_overbeam) goto ENDBEAM;  /* Can only beam over one barline */
 
-    if (out_bar < curmovt->barcount)
-      {
-      stavestr *ss = curmovt->stavetable[out_stave];
-      if (ss == NULL || ss->barindex[out_bar+1] == NULL) goto ENDBEAM;
-      beam_overbeam = TRUE;
-      overbarcount = adjust = 0;
-      lastinbar = lastp;            /* May point to a rest */
-      lastmoffinbar = lastmoff;
-      }
-    break;
+    /* The bar line has been flagged for continuation of a beam over it. We
+    arrange to keep on scanning from the start of the next bar, if there is
+    one. */
 
-    case b_Jump:
-    pp = (b_notestr *)(((b_Jumpstr *)pp)->next);
+    pp = (b_notestr *)(((b_overbeamstr *)pp)->nextbar);
+    if (pp == NULL) goto ENDBEAM;
+    beam_overbeam = TRUE;
+    overbarcount = move_adjust = 0;
+    lastinbar = lastp;            /* May point to a rest */
+    lastmoffinbar = lastmoff;
     break;
 
     case b_note:
@@ -951,9 +963,9 @@ for (;;)
       if (notetype < quaver) goto ENDBEAM;
       if (lastp->length != 0)
         currentmoff += lastp->length; else currentmoff++;
-      adjusts[adjustptr] = lastadjust = adjust + out_grace_fudge;
+      adjusts[adjustptr] = lastadjust = move_adjust + out_gracefudge;
 
-      if (pp->spitch != 0 || curmovt->beamendrests)
+      if (pp->spitch != 0 || MFLAG(mf_beamendrests))
         {
         beam_lastmoff = currentmoff;
         beam_last = pp;
@@ -971,25 +983,27 @@ for (;;)
       }
 
     adjustptr++;
-    adjust = 0;
+    move_adjust = 0;
     break;
 
+    /* For a secondary beam break, adjust longestnote to correspond to the
+    length of note implied by the number of beams that are left. */
+
     case b_beambreak:
+    bb = (b_beambreakstr *)pp;
+    if (bb->value == BEAMBREAK_ALL) goto ENDBEAM;  /* Break all */
+    if (quaver - 1 + bb->value < longestnote)
+      longestnote = quaver - 1 + bb->value;
+    break;
+
     case b_reset:
     goto ENDBEAM;
 
-    case b_beambreak2:
-      {
-      int j = quaver - 1 + ((b_beambreak2str *)pp)->value;
-      if (j < longestnote) longestnote = j;
-      }
-    break;
-
     /* Set up move amount in case the next thing is a note. We need to keep
     track of movements to get the beams moved too. */
 
     case b_move:
-    adjust = ((b_movestr *)pp)->x;
+    move_adjust = ((b_movestr *)pp)->x;
     break;
 
     /* Miscellaneous things that are permitted in the middle of beams. If it's
@@ -998,22 +1012,16 @@ for (;;)
 
     case b_dotbar: case b_clef: case b_lrepeat: case b_rrepeat:
     case b_comma: case b_caesura: case b_tick:
-    adjust = 0;
+    move_adjust = 0;
     break;
     }
-
-  /* Move on to the next item */
-
-  pp = (b_notestr *)((uschar *)pp + length_table[type]);
-  type = pp->type;
   }
 
-/* Label to jump to to get out of the switch in the loop */
+/* Label to jump to to get out of the switch in the loop. */
 
 ENDBEAM:
 
-/* Correct the count for rests that might have been beamed over, but weren't.
-*/
+/* Correct the count for rests that might have been beamed over, but weren't. */
 
 beam_count -= pendingrests;
 
@@ -1022,7 +1030,7 @@ the start of a system. In this case, a forced slope will be set. */
 
 if (beam_last == beam_first && !beam_continued)
   {
-  DEBUG(("out_setupbeam() FALSE\n"));
+  TRACE("out_setupbeam() FALSE\n");
   return FALSE;
   }
 
@@ -1033,7 +1041,7 @@ if (beam_overbeam && overbarcount <= 0)
   {
   if (nodraw)
     {
-    DEBUG(("out_setupbeam() FALSE\n"));
+    TRACE("out_setupbeam() FALSE\n");
     return FALSE;
     }
   beam_overbeam = FALSE;
@@ -1048,7 +1056,7 @@ if (beam_last->spitch == 0) lastopposite = !beam_upflag;
 /* Compute the xcorrection for any notes with stems up, and for notes on the
 opposite side of the beam. */
 
-beam_Xcorrection = mac_muldiv(51*main_stavemagn, n_fontsize, 100000);
+beam_Xcorrection = mac_muldiv(51*out_stavemagn, n_fontsize, 100000);
 
 /* Set up the coordinates of the left-hand end of the beam, relative to the
 stave base and the start of the bar (with the y coordinate going upwards). The
@@ -1056,21 +1064,22 @@ x correction of 0.75 point is required because the downward stems start that
 far from the note's origin. */
 
 beam_firstX = out_findXoffset(beam_firstmoff) + adjusts[0] +
-  mac_muldiv(n_fontsize, 75*main_stavemagn, 1000000);
+  mac_muldiv(n_fontsize, 75*out_stavemagn, 1000000);
 
 beam_firstpitch = p->spitch;
 
 if (beam_upflag && !restatstart) beam_firstX += beam_Xcorrection;
 mac_couplepitch(beam_firstpitch, p->flags);
 
-beam_firstY = (beam_firstpitch - 128 + n_upfactor*14)*main_stavemagn;
+beam_firstY = (beam_firstpitch - P_1L)*out_pitchmagn +
+  n_upfactor*14*out_stavemagn;
 
 /* This "loop", which computes a slope and an offset, is obeyed twice if on the
 first pass we fail to arrange for notes on the opposite side of the beam, and
 therefore have to do it all again. Exit should always be via "break", but code
 it with a counter, just in case. */
 
-for (i = 0; i < 2; i++)
+for (int i = 0; i < 2; i++)
   {
   int j, pitchright;
   BOOL offsetOK = FALSE;
@@ -1080,19 +1089,19 @@ for (i = 0; i < 2; i++)
   0.4pt thickness of stems. */
 
   xright = out_findXoffset(beam_lastmoff) + lastadjust +
-    mac_muldiv(n_fontsize, 115*main_stavemagn, 1000000);
+    mac_muldiv(n_fontsize, 115*out_stavemagn, 1000000);
 
   if (beam_upflag != lastopposite) xright += beam_Xcorrection;
 
   pitchright = beam_last->spitch;
   mac_couplepitch(pitchright, beam_last->flags);
-  yright = (pitchright - 128 +
-    (((beam_last->flags & nf_stemup) != 0)? 14 : -14))*main_stavemagn;
+  yright = (pitchright - P_1L)*out_pitchmagn +
+    (((beam_last->flags & nf_stemup) != 0)? 14 : -14)*out_stavemagn;
 
   /* Now we must select a slope for the beam. First check to see if something
   has been forced. */
 
-  if (beam_forceslope != BIGNUMBER) beam_slope = beam_forceslope;
+  if (beam_forceslope != INT32_MAX) beam_slope = beam_forceslope;
 
   /* If there is only one real note in the beam (can happen with leading and/or
   trailing rests) the slope is horizontal. */
@@ -1124,12 +1133,12 @@ for (i = 0; i < 2; i++)
       yright -= n_upfactor*(yleftadjust + yrightadjust);
       }
 
-    beam_slope = mac_muldiv(1000, yyright-yyleft, xright-beam_firstX);
+    beam_slope = mac_muldiv(1000, yyright - yyleft, xright - beam_firstX);
 
     /* Apply a maximum slope */
 
-    if (abs(beam_slope) > curmovt->maxbeamslope1)
-      beam_slope = (beam_slope/abs(beam_slope))*curmovt->maxbeamslope1;
+    if (abs(beam_slope) > curmovt->maxbeamslope[0])
+      beam_slope = (beam_slope/abs(beam_slope))*curmovt->maxbeamslope[0];
     }
 
   /* More than two notes in the beam. Select a slope using the heuristics in
@@ -1137,9 +1146,9 @@ for (i = 0; i < 2; i++)
 
   else
     {
-    int abs_slope;
-    int slope = findslope();
-    int yslope = beam_firstY + mac_muldiv(slope, xright-beam_firstX, 1000);
+    int32_t abs_slope;
+    int32_t slope = findslope();
+    int32_t yslope = beam_firstY + mac_muldiv(slope, xright-beam_firstX, 1000);
 
     /* If the slope supplied by the heuristic is zero, or if it is less than
     the slope of the line joining the end notes, use it. Otherwise join the end
@@ -1148,7 +1157,7 @@ for (i = 0; i < 2; i++)
     if (slope == 0 || (slope > 0 && yslope < yright) ||
       (slope < 0 && yslope > yright)) yright = yslope;
 
-    beam_slope = mac_muldiv(1000, yright-beam_firstY, xright-beam_firstX);
+    beam_slope = mac_muldiv(1000, yright - beam_firstY, xright - beam_firstX);
 
     /* When the final note is on the other side of the beam, it is helpful to
     keep the slope down to a smaller maximum to enable more cases to be fitted
@@ -1165,9 +1174,9 @@ for (i = 0; i < 2; i++)
 
   for (j = 0; j < 5; j++)
     {
-    int ornament = (n_ornament == NULL)? -1 : n_ornament->ornament;
+    //int ornament = (n_ornament == NULL)? -1 : n_ornament->ornament;
     if ((offsetOK =
-      ComputeBeamOffset(ornament, lastopposite, adjusts, longestnote))
+      ComputeBeamOffset(lastopposite, adjusts, longestnote))
         == TRUE) break;
     beam_slope = n_upfactor*retryslopes[j];
     }
@@ -1183,19 +1192,45 @@ for (i = 0; i < 2; i++)
 
     for (;;)
       {
-      if ((bfp->flags & nf_stemup) != flag) read_resetstemflag(bfp, flag);
+      if ((bfp->flags & nf_stemup) != flag)
+        {
+        b_notestr *xp = bfp;
+
+        /* Before we reset the flag for a chord, we must reset the offsets of
+        any accidentals, unless there is an explicit setting, because when the
+        stem of a chord is set, the accidental positions are calculated,
+        assuming that what is there already is the basic width of the
+        accidental. As this is an error situation, we don't have to get it
+        perfect. */
+
+        do
+          {
+          if (xp->acc != 0)
+            {
+            xp->accleft += curmovt->accspacing[xp->acc] -
+              curmovt->accadjusts[xp->notetype];
+            if ((xp->flags & (nf_accrbra+nf_accsbra)) != 0)
+              xp->accleft += (xp->acc == ac_df)? 6800 : 5300;
+            }
+          xp->flags &= ~nf_stemup;
+          xp = (b_notestr *)(xp->next);
+          }
+        while (xp->type == b_chord);
+        mac_setstemflag(bfp, flag);
+        }
+
       if (bfp == beam_last) break;
-      bfp = misc_nextnote(bfp, NULL);
+      bfp = misc_nextnote(bfp);
       }
 
-    error_moan(ERR61, out_bar, out_stave);
+    error(ERR143);
     lastopposite = FALSE;
     }
-  }
+  }   /* End of loop to retry on failure */
 
 /* Reset the forced slope */
 
-beam_forceslope = BIGNUMBER;
+beam_forceslope = INT32_MAX;
 
 /* Incorporate the computed and manual adjustment offset into the first y
 position. */
@@ -1218,7 +1253,7 @@ left is drawn. */
 
 if (beam_continued)
   {
-  int beamadjust = 4 * main_stavemagn;
+  int32_t beamadjust = 4 * out_stavemagn;
   if (beam_upflag) beamadjust += beam_Xcorrection;
   beam_firstX -= beamadjust;
   beam_firstY -= (beam_slope * beamadjust) / 1000;
@@ -1250,9 +1285,9 @@ if (beam_overbeam)
   {
   obeamstr *b;
   if (bar_cont->overbeam == NULL)
-    bar_cont->overbeam = store_Xget(sizeof(obeamstr));
-  b = bar_cont->overbeam;
+    bar_cont->overbeam = mem_get(sizeof(obeamstr));
 
+  b = bar_cont->overbeam;
   b->firstX = beam_firstX - out_poslast->xoff - out_sysblock->barlinewidth;
   b->firstY = beam_firstY;
   b->slope = beam_slope;
@@ -1261,13 +1296,12 @@ if (beam_overbeam)
   b->Xcorrection = beam_Xcorrection;
   b->splitOK = beam_splitOK;
   b->upflag = beam_upflag;
-
   beam_overbeam = FALSE;
   }
 
 /* A yield of TRUE indicates that beaming is now in force */
 
-DEBUG(("out_setupbeam() TRUE\n"));
+TRACE("out_setupbeam() TRUE\n");
 return TRUE;
 }
 
