@@ -1,19 +1,12 @@
 /*************************************************
-*    The PMW Music Typesetter - 3rd incarnation  *
+*            PMW code for setting slurs          *
 *************************************************/
 
-/* Copyright (c) Philip Hazel, 1991 - 2020 */
+/* Copyright Philip Hazel 2021 */
+/* This file created: July 2021 */
+/* This file last modified: July 2021 */
 
-/* Written by Philip Hazel, starting November 1991 */
-/* This file last modified: April 2020 */
-
-
-/* This file contains code for miscellanous functions used
-for drawing slurs. */
-
-#include "pmwhdr.h"
-#include "outhdr.h"
-#include "pagehdr.h"
+#include "pmw.h"
 
 
 
@@ -21,35 +14,34 @@ for drawing slurs. */
 *         Set up start of slur processing        *
 *************************************************/
 
-/* This is called when processing [slur], both while setting up the cont
-structure, and while actually outputting.
+/* This is called when processing [slur], both while paginating and also while
+actually outputting. Active slurs are held on a chain; when the slur is
+complete, the block is added to a free chain so that it can be re-used.
 
-Argument:  the slur start item
-Returns:   the slur processing structure that was set up
+Argument:   the slur start item
+Returns:    the slur processing structure that was set up
 */
 
 slurstr *
-misc_setstartslur(bstr *p)
+slur_startslur(b_slurstr *p)
 {
-b_slurstr *pp = (b_slurstr *)p;
-slurstr *s = store_Xget(sizeof(slurstr));
+slurstr *s = mem_get_cached((void **)(&main_freeslurblocks), sizeof(slurstr));
 
-s->slur = pp;
-s->maxy = -BIGNUMBER;
-s->miny =  BIGNUMBER;
+s->slur = p;
+s->maxy = INT32_MIN;
+s->miny = INT32_MAX;
 s->gaps = NULL;
 s->section = 1;
-
 s->lastx = s->x = s->count = s->slopeleft = s->sloperight = 0;
 
-/* Put in a default y value in case the slur doesn't cross anything visible */
+/* Put in a default y value in case the slur doesn't cross anything visible. */
 
-s->lasty = s->y = ((pp->flags & sflag_b) == 0)? 16000 : 0;
+s->lasty = s->y = ((p->flags & sflag_b) == 0)? 16000 : 0;
 
 /* If this was a crossing slur, we place it second on the stack, if possible.
-Otherwise put it at th top. */
+Otherwise put it at the top. */
 
-if ((pp->flags & sflag_x) != 0 && bar_cont->slurs != NULL)
+if ((p->flags & sflag_x) != 0 && bar_cont->slurs != NULL)
   {
   s->next = (bar_cont->slurs)->next;
   (bar_cont->slurs)->next = s;
@@ -63,49 +55,52 @@ return s;
 }
 
 
+
 /*************************************************
 *        Find and unchain given slur             *
 *************************************************/
 
-/* This is called when processing [endslur], both while setting up the cont
-structure, and while actually outputting. If [endslur] has an id, we seek that
-slur, yielding NULL if not found; otherwise the first on the chain is yielded.
+/* This is called when processing [endslur] or [endline], both while setting up
+the cont structure, and while actually outputting. A slur will always be in
+progress; an isolated [endslur] is detected at reading time, causing an error
+that stops further processing. If [endslur] has an id, we seek that slur,
+yielding NULL if not found; otherwise the first on the chain is yielded.
 
-Argument:  the endslur item
-Returns:   the found slur structure, removed from the chain of active slurs
+Argument:   the slur end item
+Returns:    the active slur block or NULL if not found
 */
 
 slurstr *
-misc_getendslur(bstr *p)
+slur_endslur(b_endslurstr *p)
 {
 slurstr *s = bar_cont->slurs;
-int slurid = ((b_endslurstr *)p)->id;
+int slurid = p->value;
 
-if (slurid != 0)
+if (slurid == 0) bar_cont->slurs = s->next; else
   {
   slurstr **ss = &(bar_cont->slurs);
-  BOOL found = FALSE;
-
   while (s != NULL)
     {
     if ((s->slur)->id == slurid)
       {
-      found = TRUE;
       *ss = s->next;
       break;
       }
     ss = &(s->next);
     s = *ss;
     }
-
-  if (!found) return NULL;
+  if (s == NULL) return NULL;
   }
 
-else bar_cont->slurs = s->next;
+/* We can put the slur block back onto its free chain, because its contents
+will be used straight away, before any subsequent [slur] is processed. */
+
+mem_free_cached((void **)&main_freeslurblocks, s);
 return s;
 }
 
 
+
 /*************************************************
 *         Find Bezier parameter fraction         *
 *************************************************/
@@ -196,14 +191,14 @@ Arguments:
 Returns:     vector of left, middle, right coordinates
 */
 
-static int coords[6];
+static int32_t coords[6];
 
 static int *
-getgapcoords(int ix0, int iy0, int ix1, int iy1, int flags, int co, int start,
-  int stop)
+getgapcoords(int32_t ix0, int32_t iy0, int32_t ix1, int32_t iy1, uint32_t flags,
+  int32_t co, int32_t start, int32_t stop)
 {
 int above, wiggly;
-int ox, oy;
+int32_t ox, oy;
 double xx, yy, w, v;
 double x0, x1, x2, x3, y0, y1, y2, y3;
 double a, b, c, f, g;
@@ -225,7 +220,7 @@ w /= 2.0;
 v = w*0.6667;
 
 if (v > 10000.0) v = 10000.0;
-co = (above * (co + ((xx > 20000)? 6000 : (int)(xx * 0.3))) * main_stavemagn)/1000;
+co = (above * (co + ((xx > 20000)? 6000 : (int)(xx * 0.3))) * out_stavemagn)/1000;
 
 f = ((double)start)/1000.0;
 g = ((double)stop)/1000.0;
@@ -234,7 +229,7 @@ g = ((double)stop)/1000.0;
 We don't actually have to translate or rotate, since we are not actually going
 to draw anything here. */
 
-ox = (ix0+ix1+6*main_stavemagn)/2;
+ox = (ix0+ix1+6*out_stavemagn)/2;
 oy = (iy0+iy1)/2;
 
 /* Set up traditional Bezier coordinates for the complete slur. */
@@ -294,6 +289,7 @@ return coords;
 }
 
 
+
 /*************************************************
 *      Output text in a slur or line gap         *
 *************************************************/
@@ -308,87 +304,93 @@ Returns:       nothing
 */
 
 static void
-out_gaptext(gaptextstr *gt, int x, int y, int num, int den)
+slur_gaptext(b_slurgapstr *sg, int32_t x, int32_t y, int32_t num, int32_t den)
 {
-int *matrix;
-int width, sn, cs;
-int unscaled_fontsize = ((curmovt->fontsizes)->fontsize_text)[gt->size];
-int fontsize = mac_muldiv(main_stavemagn, unscaled_fontsize, 1000);
-
-matrix = ((curmovt->fontsizes)->fontmatrix_text)[gt->size];
-if (matrix != NULL) memcpy(font_transform, matrix, 4*sizeof(int));
+int32_t adjusthy, width, sn, cs;
+fontinststr localfdata, *fdata;
+fontinststr *origfdata = curmovt->fontsizes->fontsize_text + sg->textsize;
 
 if (num != 0)    /* Not horizontal */
   {
   double hy = sqrt((double)num*(double)num + (double)den*(double)den);
-  font_rotate((int)((45000.0*atan2((double)num, (double)den))/atan(1.0)));
+  fdata = font_rotate(origfdata,
+    (int)((45000.0*atan2((double)num, (double)den))/atan(1.0)));
   sn = (int)((1000.0*(double)num)/hy);
   cs = (int)((1000.0*(double)den)/hy);
   }
 else
   {
+  localfdata = *origfdata;
+  fdata = &localfdata;
   sn = 0;
   cs = 1000;
   den = 1;   /* To stop division by 0 below */
   }
 
-width = string_width(gt->text, font_rm, fontsize);
+fdata->size = mac_muldiv(fdata->size, out_stavemagn, 1000);
+adjusthy = sg->texty - fdata->size/4;
+width = string_width(sg->gaptext, fdata, NULL);
 
 x = x - width/2 +
-  mac_muldiv(gt->x, cs, 1000) - mac_muldiv(gt->y - fontsize/4, sn, 1000);
+  mac_muldiv(sg->textx, cs, 1000) - mac_muldiv(adjusthy, sn, 1000);
 
 y = y - (width*num)/(2*den) +
-  mac_muldiv(gt->x, sn, 1000) + mac_muldiv(gt->y - fontsize/4, cs, 1000);
+  mac_muldiv(sg->textx, sn, 1000) + mac_muldiv(adjusthy, cs, 1000);
 
-out_string(gt->text, font_rm, fontsize, x, out_ystave - y, gt->flags);
-font_reset();
+out_string(sg->gaptext, fdata, x, out_ystave - y, sg->textflags);
 }
 
 
+
 /*************************************************
 *   Compute parameters & draw a slur or line     *
 *************************************************/
 
-/* This function is called when [endslur] is reached, or at the end of a line.
+/* This function is called when [endslur] or [endline] is reached or if the end
+of a line is reached in the middle of a slur. The active slur structure s has
+already been placed at the head of the chain of free slur blocks, knowing that
+its data will be used before any subsequent [slur] is processed, so we do not
+need to dispose of it here.
 
 Arguments:
   s          the active slur structure
   x1         the x coordinate of the end of the slur
-  npitch     pitch of the final note
+  npitch     pitch of the final note or 0 if drawing to end of line
   eol        TRUE if this slur goes to the end of the line
 
 Returns:     nothing
 */
 
 void
-misc_drawslur(slurstr *s, int x1, int npitch, BOOL eol)
+slur_drawslur(slurstr *s, int x1, int npitch, BOOL eol)
 {
 b_slurstr *ss = s->slur;
 b_slurmodstr *sm = NULL;
 b_slurmodstr *smm = ss->mods;
 
-BOOL sol;
-BOOL laststemup = out_laststemup[out_stave];
+uint32_t slurflags = ss->flags;
 
-int adjustco = 0;
-int slurflags = ss->flags;
-int lineslur = (slurflags & sflag_l) != 0;
-int below = (slurflags & sflag_b) != 0;
-int absolute = (slurflags & sflag_abs) != 0;
-int lay = (slurflags & sflag_lay) != 0;
-int use_endmoff = ((slurflags & sflag_cx) != 0)? out_moff : out_lastmoff;
-int line_rc_adjust = lineslur? 3000 : 0;
+BOOL absolute = (slurflags & sflag_abs) != 0;
+BOOL below = (slurflags & sflag_b) != 0;
+BOOL lay = (slurflags & sflag_lay) != 0;
+BOOL laststemup = out_laststemup[curstave];
+BOOL lineslur = (slurflags & sflag_l) != 0;
 
-int x0 = s->x;
-int y0 = s->y;
-int y1;
+int32_t adjustco = 0;
+int32_t use_endmoff = ((slurflags & sflag_cx) != 0)? out_moff : out_lastmoff;
+int32_t line_rc_adjust = lineslur? 3000 : 0;
+int32_t x0 = s->x;
+int32_t y0 = s->y;
+int32_t y1;
+
+BOOL sol = x0 == 0;
 
 /* For continued slurs in style 0 we have to adjust the apparent end points and
 then draw only part of the resulting slur. These are the start and end
 fractional values. */
 
-int dstart = 0;
-int dstop = 1000;
+int32_t dstart = 0;
+int32_t dstop = 1000;
 
 /* Note: for absolute and {und,ov}erlay positioning, all this computation is
 later over-ridden. We leave it in place so as not to disturb things for the x
@@ -409,14 +411,16 @@ if (eol)
   if (x0 != 0) slurflags &= ~sflag_w;   /* No wiggle */
   slurflags |= sflag_or;                /* Open on right if line */
   }
-else    /* Not end of line slur */
-  {
-  y1 = (npitch == 0)? L_2L :
-    misc_ybound(below, n_prevtie, TRUE, TRUE);
 
-  /* If the last note was beamed, and the slur is on the same side as the beam,
-  we need to put in an additional bit of space for clearance. Also, if the slur
-  is on the opposite side to the stem, ditto. */
+/* Not end of line slur. Set the vertical position according to the last note.
+If the last note was beamed, and the slur is on the same side as the beam, or
+if the slur is on the opposite side to the stem, we need to put in an
+additional bit of space for clearance. */
+
+else
+  {
+  y1 = (npitch == 0)? (below? L_2L : L_5L) :
+    misc_ybound(below, n_prevtie, TRUE, TRUE);
 
   if (below)
     {
@@ -436,15 +440,13 @@ final part of a split wiggly slur. */
 
 if (x0 == 0)
   {
-  sol = TRUE;
   slurflags |= sflag_ol;              /* Open on left if line */
   if ((slurflags & sflag_w) != 0)
     {
     slurflags &= ~sflag_w;            /* No wiggle */
     below = !below;                   /* Other direction */
     slurflags ^= sflag_b;             /* Must do this too */
-    y1 = (npitch == 0)? L_2L :
-      misc_ybound(below, n_prevtie, TRUE, TRUE);
+    y1 = (npitch == 0)? L_2L : misc_ybound(below, n_prevtie, TRUE, TRUE);
     }
 
   x0 = out_sysblock->firstnoteposition + out_sysblock->xjustify - 10500;
@@ -455,17 +457,15 @@ if (x0 == 0)
     if (lineslur) x0 -= 10000;   /* lines normally start 10 pts right */
     }
   }
-else sol = FALSE;
 
 /* For wiggly slurs, move the final point to the other end of the last note. We
 don't attempt any other fancy adjustments for these slurs. */
 
 if ((slurflags & sflag_w) != 0)
   {
-  y1 = (npitch == 0)? L_2L :
-    misc_ybound(!below, n_prevtie, TRUE, TRUE);
+  y1 = (npitch == 0)? L_2L : misc_ybound(!below, n_prevtie, TRUE, TRUE);
   if (!below && !laststemup && (n_flags & nf_stem) != 0)
-    x1 -= 5*main_stavemagn;
+    x1 -= 5*out_stavemagn;
   }
 
 /* For non-wiggly slurs, make adjustments according to the starting and ending
@@ -476,14 +476,14 @@ under the slur. */
 
 else if (below)    /* slur below */
   {
-  int miny = s->miny;
+  int32_t miny = s->miny;
 
   if (!laststemup &&               /* Note has down stem */
       (n_flags & nf_stem) != 0 &&  /* Note has a stem */
       !lineslur &&                 /* Not a line (i.e. a slur) */
       !eol &&                      /* Not an end of line slur */
       !main_righttoleft)           /* Not right-to-left */
-    x1 -= 5*main_stavemagn;
+    x1 -= 5*out_stavemagn;
 
   if (s->count > 3)
     {
@@ -495,7 +495,7 @@ else if (below)    /* slur below */
 
   if (miny < y0 && miny < y1)
     {
-    int adjust = (y0 + y1)/2 - miny;
+    int32_t adjust = (y0 + y1)/2 - miny;
     if (lineslur)
       {
       if ((slurflags & sflag_h) != 0) y0 = y1 = miny;
@@ -507,7 +507,7 @@ else if (below)    /* slur below */
 
 else    /* slur above */
   {
-  int maxy = s->maxy;
+  int32_t maxy = s->maxy;
   if (s->count > 3)
     {
     if (s->slopeleft > 400)
@@ -533,8 +533,7 @@ other cases not yet). */
 
 if ((slurflags & sflag_h) != 0)
   {
-  if ((!below && y1 > y0) || (below && y1 < y0)) y0 = y1;
-    else y1 = y0;
+  if ((!below && y1 > y0) || (below && y1 < y0)) y0 = y1; else y1 = y0;
   }
 
 /* If this is a curved slur, arrange that the end points are not on staff
@@ -544,8 +543,8 @@ the slur is short. */
 
 if (!lineslur)
   {
-  int ay0 = abs(y0);
-  int ay1 = abs(y1);
+  int32_t ay0 = abs(y0);
+  int32_t ay1 = abs(y1);
 
   if (below)
     {
@@ -562,7 +561,7 @@ if (!lineslur)
 
   else if (!absolute && !lay && x1 - x0 > 24000)
     {
-    int overall_slope = mac_muldiv(y1 - y0, 1000, x1 - x0);
+    int32_t overall_slope = mac_muldiv(y1 - y0, 1000, x1 - x0);
     if (abs(overall_slope) > 500 && adjustco < 2000) adjustco = 2000;
     }
   }
@@ -597,7 +596,7 @@ if (absolute)
 else if (lay)
   {
   y0 = y1 = below?
-    out_sysblock->ulevel[out_stave] : out_sysblock->olevel[out_stave];
+    out_sysblock->ulevel[curstave] : out_sysblock->olevel[curstave];
   }
 
 /* Finally, apply manual adjustments. All endpoints of all sections are
@@ -624,7 +623,7 @@ if (sol || eol)
     {
     if (sm->lxoffset != 0)
       {
-      int offset = mac_muldiv(len_crotchet, sm->lxoffset, 1000);
+      int32_t offset = mac_muldiv(len_crotchet, sm->lxoffset, 1000);
       x0 = out_barx + out_findAoffset(s->moff + offset);
       }
 
@@ -633,7 +632,7 @@ if (sol || eol)
 
     if (sm->rxoffset != 0)
       {
-      int offset = mac_muldiv(len_crotchet, sm->rxoffset, 1000);
+      int32_t offset = mac_muldiv(len_crotchet, sm->rxoffset, 1000);
       x1 = out_barx + out_findAoffset(use_endmoff + offset) - line_rc_adjust;
       }
     x0 += sm->lx;
@@ -641,10 +640,10 @@ if (sol || eol)
     y0 += sm->ly;
     y1 += sm->ry;
     adjustco += sm->c;
-    out_slurclx = (sm->clx * main_stavemagn)/1000;
-    out_slurcly = (sm->cly * main_stavemagn)/1000;
-    out_slurcrx = (sm->crx * main_stavemagn)/1000;
-    out_slurcry = (sm->cry * main_stavemagn)/1000;
+    out_slurclx = (sm->clx * out_stavemagn)/1000;
+    out_slurcly = (sm->cly * out_stavemagn)/1000;
+    out_slurcrx = (sm->crx * out_stavemagn)/1000;
+    out_slurcry = (sm->cry * out_stavemagn)/1000;
     }
   }
 
@@ -667,7 +666,7 @@ else if (sol)
     {
     if (smm->rxoffset != 0)
       {
-      int offset = mac_muldiv(len_crotchet, smm->rxoffset, 1000);
+      int32_t offset = mac_muldiv(len_crotchet, smm->rxoffset, 1000);
       x1 = out_barx + out_findAoffset(use_endmoff + offset) - line_rc_adjust;
       }
     x1 += smm->rx;
@@ -686,7 +685,7 @@ else if (eol)
     {
     if (smm->lxoffset != 0)   /* Relative to start of slur */
       {
-      int offset = mac_muldiv(len_crotchet, smm->lxoffset, 1000);
+      int32_t offset = mac_muldiv(len_crotchet, smm->lxoffset, 1000);
       x0 = out_barx + out_findAoffset(s->moff + offset);
       }
     x0 += smm->lx;
@@ -701,12 +700,12 @@ else if (smm != NULL)
   {
   if (smm->lxoffset != 0)  /* Relative to start of slur */
     {
-    int offset = mac_muldiv(len_crotchet, smm->lxoffset, 1000);
+    int32_t offset = mac_muldiv(len_crotchet, smm->lxoffset, 1000);
     x0 = out_barx + out_findAoffset(s->moff + offset);
     }
   if (smm->rxoffset != 0)
     {
-    int offset = mac_muldiv(len_crotchet, smm->rxoffset, 1000);
+    int32_t offset = mac_muldiv(len_crotchet, smm->rxoffset, 1000);
     x1 = out_barx + out_findAoffset(use_endmoff + offset) - line_rc_adjust;
     }
   x0 += smm->lx;
@@ -714,10 +713,10 @@ else if (smm != NULL)
   y0 += smm->ly;
   y1 += smm->ry;
   adjustco += smm->c;
-  out_slurclx = (smm->clx * main_stavemagn)/1000;
-  out_slurcly = (smm->cly * main_stavemagn)/1000;
-  out_slurcrx = (smm->crx * main_stavemagn)/1000;
-  out_slurcry = (smm->cry * main_stavemagn)/1000;
+  out_slurclx = (smm->clx * out_stavemagn)/1000;
+  out_slurcly = (smm->cly * out_stavemagn)/1000;
+  out_slurcrx = (smm->crx * out_stavemagn)/1000;
+  out_slurcry = (smm->cry * out_stavemagn)/1000;
   }
 
 /* Need to correct for the jog length for absolute line slurs, so that the line
@@ -725,7 +724,7 @@ is at the height specified. Also need to move above slurs up. */
 
 if (lineslur && (absolute || lay))
   {
-  int x = adjustco + 3000;
+  int32_t x = adjustco + 3000;
   if (!below) x = -x;
   y0 += x;
   y1 += x;
@@ -733,50 +732,70 @@ if (lineslur && (absolute || lay))
 
 /* Adjust the verticals for the stave magnification */
 
-y0 = mac_muldiv(y0, main_stavemagn, 1000);
-y1 = mac_muldiv(y1, main_stavemagn, 1000);
+y0 = mac_muldiv(y0, out_stavemagn, 1000);
+y1 = mac_muldiv(y1, out_stavemagn, 1000);
 
 /* Make adjustments and fudges for continued curved slurs if the style is not
-the default. */
+the default. We do this by drawing only half a slur. Such slurs may be
+continued at either end. */
 
-if (eol && !lineslur && curmovt->endlineslurstyle != 0)
+if (!lineslur && curmovt->endlineslurstyle != 0 && (sol || eol))
   {
-  x1 += x1 - x0;
-  y1 += (y1 - y0)/5;
-  adjustco += 2*(y1 - y0)/3;
-  dstop = 500;
+  int32_t dx = x1 - x0;
+  int32_t dy = y1 - y0;
+
+  /* If this is both as starting and ending part (hopefully a rare occurrence),
+  make the slur a bit longer and draw from 0.1 to 0.9 of its length. */
+
+  if (sol && eol)
+    {
+    x0 -= dx/8;
+    x1 += dx/8;
+    dstart = 100;
+    dstop = 900;
+    }
+
+  /* If this is a starting part or an ending part, adjust the values so that
+  we draw exactly half a slur, which will have the desired look. */
+
+  else
+    {
+    adjustco += 4*dy/5;
+
+    if (sol)
+      {
+      x0 -= dx;
+      y0 -= dy/5;
+      dstart = 500;
+      }
+    else
+      {
+      x1 += dx;
+      y1 += dy/5;
+      dstop = 500;
+      }
+    }
   }
 
-if (sol && !lineslur && curmovt->endlineslurstyle != 0)
-  {
-  x0 -= x1 - x0;
-  y0 -= (y1 - y0)/5;
-  adjustco += 2*(y1 - y0)/3;
-  dstart = 500;
-  }
+/* There may be a chain of gaps in a line or slur. Note that the out_slur()
+function adds 7*out_stavemagn to the x1 value, to get it past the final note,
+so we have to compensate for that here. */
 
-/* If this is a line slur, there may be a chain of gaps to be dealt with. Note
-that the out_slur routine adds 7*main_stavemagn to the x1 value, to get it past
-the final note. */
-
-if (lineslur && s->gaps != NULL)
+if (s->gaps != NULL)
   {
-  b_linegapstr *lg;
-  gapstr *pg = s->gaps;
-  int fudge = 7*main_stavemagn;
+  BOOL done = FALSE;
+  int32_t fudge = 7*out_stavemagn;
+  int32_t xlength = x1 + fudge - x0;
+  int32_t start = dstart;
 
   /* First scan through and compute positions for any that are specified as a
-  fraction of the way along the line. */
+  fraction of the way along. */
 
-  while (pg != NULL)
+  for (gapstr *pg = s->gaps; pg != NULL; pg = pg->next)
     {
-    lg = pg->gap;
-    if (lg->hfraction >= 0)
-      {
-      pg->x = lg->xadjust + x0 +
-        mac_muldiv(x1 + fudge - x0, lg->hfraction, 1000);
-      }
-    pg = pg->next;
+    int hfraction = pg->gap->hfraction;
+    if (hfraction >= 0)
+      pg->x = pg->gap->xadjust + x0 + mac_muldiv(xlength, hfraction, 1000);
     }
 
   /* The gaps may be in any order, horizontally. We could sort them before
@@ -785,15 +804,14 @@ if (lineslur && s->gaps != NULL)
 
   while (s->gaps != NULL)
     {
+    b_slurgapstr *sg;
     gapstr **gg = &(s->gaps);      /* Parent ptr */
     gapstr *g = *gg;               /* Active block ptr */
-    int firstx = g->x;
-    int xg0, yg0, num, den, xwidth;
+    int32_t firstx = g->x;
 
     /* Scan to find the leftmost */
 
-    pg = g;
-    while (pg->next != NULL)
+    for (gapstr *pg = g; pg->next != NULL; pg = pg->next)
       {
       if ((pg->next)->x < firstx)
         {
@@ -801,181 +819,137 @@ if (lineslur && s->gaps != NULL)
         g = *gg;
         firstx = g->x;
         }
-      pg = pg->next;
       }
+    sg = g->gap;
 
-    /* Draw the line to this gap, unless it has negative length. */
+    /* Process a gap in a line. The values of the starting x/y coordinates are
+    adjusted for each part of the line - the start/stop parameters of
+    out_slur() are not used when drawing lines. */
 
-    num = y1 - y0;
-    den = x1 + fudge - x0;
-    lg = g->gap;
-
-    if (num == 0) /* Optimize out the common case (horizontal) */
+    if (lineslur)
       {
-      xwidth = lg->width;
-      xg0 = g->x - xwidth/2;
-      yg0 = y0;
+      int32_t xg0, yg0, xwidth;
+      int32_t num = y1 - y0;
+      int32_t den = x1 + fudge - x0;;
+
+      /* Compute end coordinates for the part of the line up to this gap. */
+
+      if (num == 0) /* Optimize the common case (horizontal) */
+        {
+        xwidth = sg->width;
+        xg0 = g->x - xwidth/2;
+        yg0 = y0;
+        }
+      else
+        {
+        double dnum = (double)num;
+        double dden = (double)den;
+        xwidth = (int)(((double)(sg->width) * dden)/sqrt(dnum*dnum + dden*dden));
+        xg0 = g->x - xwidth/2;
+        yg0 = y0 + mac_muldiv(xg0 - x0, num, den);
+        }
+
+      /* Draw the line to this gap, unless it has negative length. */
+
+      if (x0 < xg0)
+        out_slur(x0, y0, xg0 - fudge, yg0, slurflags|sflag_or, adjustco, 0, 0);
+
+      /* Update the starting position for the next section. */
+
+      x0 = xg0 + xwidth;
+      y0 = yg0 + mac_muldiv(x0 - xg0, num, den);
+      slurflags |= sflag_ol;
+      if (x0 >= x1) done = TRUE;  /* No final bit at the end */
+
+      /* If there is an associated draw function, set up the coordinates and
+      call it. Note that lines are always drawn 3 points above or below the
+      given y value, to leave space for the jog. */
+
+      if (sg->drawing != NULL)
+        {
+        draw_lgx = xwidth / 2;
+        draw_lgy = (y0 - yg0)/2;
+        draw_ox = xg0 + draw_lgx;
+        draw_oy = yg0 + draw_lgy + (below? (-3000) : (3000));
+        draw_gap = below? -1000 : +1000;
+        out_dodraw(sg->drawing, sg->drawargs, FALSE);
+        draw_lgx = draw_lgy = draw_gap = 0;
+        }
+
+      /* If there is an associated text string, arrange to print it centred
+      in the gap. */
+
+      if (sg->gaptext != NULL)
+        {
+        slur_gaptext(sg, (xg0 + x0)/2,
+          (y0 + yg0)/2 + (below? (-3000) : (3000)), num, den);
+        }
       }
+
+    /* Process a gap in a curved slur. In this case, the call to out_slur()
+    always has the original x/y coordinates, but uses the start/stop values to
+    control which part is to be drawn. */
+
     else
       {
-      double dnum = (double)num;
-      double dden = (double)den;
-      xwidth = (int)(((double)(lg->width) * dden) /
-        sqrt(dnum*dnum + dden*dden));
-      xg0 = g->x - xwidth/2;
-      yg0 = y0 + mac_muldiv(xg0 - x0, num, den);
-      }
+      int32_t stop = mac_muldiv(g->x - sg->width/2 - x0, 1000, xlength);
 
-    if (x0 < xg0)
-      out_slur(x0, y0, xg0 - fudge, yg0, slurflags | sflag_or, adjustco, 0, 1000);
+      /* Draw the current piece if it has a positive length */
 
-    /* Update the starting position for the next section. */
+      if (stop > start)
+        out_slur(x0, y0, x1, y1, slurflags, adjustco, start, stop);
 
-    x0 = xg0 + xwidth;
-    y0 = yg0 + mac_muldiv(x0 - xg0, num, den);
-    slurflags |= sflag_ol;
+      /* Compute start of next piece */
 
-    /* If there is an associated draw function, set up the coordinates and call
-    it. Note that lines are always drawn 3 points above or below the given y
-    value, to leave space for the jog. */
+      start =  mac_muldiv(g->x + sg->width/2 - x0, 1000, xlength);
+      if (start >= 1000) done = TRUE;  /* No final bit at the end */
 
-    if (lg->draw != NULL)
-      {
-      draw_lgx = xwidth / 2;
-      draw_lgy = (y0 - yg0)/2;
-      draw_ox = xg0 + draw_lgx;
-      draw_oy = yg0 + draw_lgy + (below? (-3000) : (3000));
-      draw_gap = below? -1000 : +1000;
-      out_dodraw(lg->draw, lg->args, FALSE);
-      draw_lgx = draw_lgy = draw_gap = 0;
-      }
+      /* If there is an associated draw function, set up the coordinates and
+      call it. */
 
-    /* If there is an associated text string, arrange to print it centred
-    in the gap. */
-
-    if (lg->gaptext != NULL)
-      {
-      out_gaptext(lg->gaptext, (xg0 + x0)/2,
-        (y0 + yg0)/2 + (below? (-3000) : (3000)), num, den);
-      }
-
-    /* Extract from the chain and free the block */
-
-    *gg = g->next;
-    store_free(g);
-    }
-
-  /* Now draw the rest of the line, if there is any left. */
-
-  if (x0 < x1) out_slur(x0, y0, x1, y1, slurflags, adjustco, 0, 1000);
-  }
-
-/* The ability to have gaps in curved slurs was added at a later stage. Rather
-than mess with the above code, put in separate code to handle it. The data is
-the same as for line gaps. */
-
-else if (s->gaps != NULL)
-  {
-  b_linegapstr *lg;
-  gapstr *pg = s->gaps;
-  int xlength = x1 + 7*main_stavemagn - x0;
-  int start = dstart;
-  int stop;
-
-  /* First scan through and compute positions for any that are specified as a
-  fraction of the way along the line. */
-
-  while (pg != NULL)
-    {
-    lg = pg->gap;
-    if (lg->hfraction >= 0)
-      {
-      pg->x = x0 + mac_muldiv(xlength, lg->hfraction, 1000) + lg->xadjust;
-      }
-    pg = pg->next;
-    }
-
-  /* The gaps may be in any order, horizontally. We could sort them before
-  processing, but it is just as easy to pick out the left- most and process it
-  until there are none left. */
-
-  while (s->gaps != NULL)
-    {
-    gapstr **gg = &(s->gaps);      /* Parent ptr */
-    gapstr *g = *gg;               /* Active block ptr */
-    int firstx = g->x;
-
-    /* Scan to find the leftmost */
-
-    pg = g;
-    while (pg->next != NULL)
-      {
-      if ((pg->next)->x < firstx)
+      if (sg->drawing != NULL)
         {
-        gg = &(pg->next);
-        g = *gg;
-        firstx = g->x;
+        int *c = getgapcoords(x0, y0, x1, y1, slurflags, adjustco, stop, start);
+        draw_ox = c[2];
+        draw_oy = c[3];
+        draw_lgx = c[2] - c[0];
+        draw_lgy = c[3] - c[1];
+        draw_gap = below? -1000 : +1000;
+        out_dodraw(sg->drawing, sg->drawargs, FALSE);
+        draw_lgx = draw_lgy = draw_gap = 0;
+        }
+
+      /* If there's associated text, output it. */
+
+      if (sg->gaptext != NULL)
+        {
+        int32_t *c = getgapcoords(x0, y0, x1, y1, slurflags, adjustco, stop, start);
+        slur_gaptext(sg, c[2], c[3], c[5] - c[1], c[4] - c[0]);
         }
-      pg = pg->next;
       }
 
-    /* Draw the slur to this gap, unless it has negative length. */
-
-    lg = g->gap;
-    stop = mac_muldiv(g->x - lg->width/2 - x0, 1000, xlength);
-
-    if (stop > start)
-      out_slur(x0, y0, x1, y1, slurflags, adjustco, start, stop);
-
-    /* Compute start of next piece */
-
-    start =  mac_muldiv(g->x + lg->width/2 - x0, 1000, xlength);
-
-    /* If there is an associated draw function, set up the coordinates and call
-    it. */
-
-    if (lg->draw != NULL)
-      {
-      int *c = getgapcoords(x0, y0, x1, y1, slurflags, adjustco, stop, start);
-      draw_ox = c[2];
-      draw_oy = c[3];
-      draw_lgx = c[2] - c[0];
-      draw_lgy = c[3] - c[1];
-      draw_gap = below? -1000 : +1000;
-      out_dodraw(lg->draw, lg->args, FALSE);
-      draw_lgx = draw_lgy = draw_gap = 0;
-      }
-
-    /* If there's associated text, output it. */
-
-    if (lg->gaptext != NULL)
-      {
-      int *c = getgapcoords(x0, y0, x1, y1, slurflags, adjustco, stop, start);
-      out_gaptext(lg->gaptext, c[2], c[3], c[5]-c[1], c[4]-c[0]);
-      }
-
-    /* Extract from the chain and free the block */
+    /* Extract the block from the chain and put it on its free chain. */
 
     *gg = g->next;
-    store_free(g);
-    }
+    mem_free_cached((void **)&main_freegapblocks, g);
+    }   /* End loop through gap blocks */
 
-  /* Now draw the rest of the slur, if there is any left. */
+  /* Draw the rest of the line or slur, if there is any left. */
 
-  if (start < 1000) out_slur(x0, y0, x1, y1, slurflags, adjustco, start, dstop);
+  if (!done) out_slur(x0, y0, x1, y1, slurflags, adjustco, start, dstop);
   }
 
-/* Else we have a slur or line with no gaps specified. Output it, provided it
-has some positive horizontal length. If there isn't enough length, generate an
-error and invent some suitable length. */
+/* The slur or line has no gaps specified. Output it, provided it has some
+positive horizontal length. If there isn't enough length, generate an error and
+invent some suitable length. */
 
 else
   {
   if (x0 >= x1)
     {
-    error_moan(ERR63, out_bar, out_stave);
+    error(ERR144);
     if (x0 == x1) x1 = x0 + 10000; else
-      { int temp = x0; x0 = x1; x1 = temp; }
+      { int32_t temp = x0; x0 = x1; x1 = temp; }
     }
   out_slur(x0, y0, x1, y1, slurflags, adjustco, dstart, dstop);
   }
@@ -983,13 +957,10 @@ else
 /* Reset the globals that hold control point adjustments. */
 
 out_slurclx = out_slurcly = out_slurcrx = out_slurcry = 0;
-
-/* Free the slur block, which is now finished with. */
-
-store_free(s);
 }
 
 
+
 /*************************************************
 *      Draw slur or line from given values       *
 *************************************************/
@@ -1013,8 +984,8 @@ parameter values in the global variables out_slurclx, out_slurcly, out_slurcrx,
 and out_slurcry for corrections to the control points.
 
 The t-values are the Bezier parameter values for drawing part slurs, given as
-fixed point values between 0 and 1.0 respectively. For a whole slur, their int
-values are therefore 0 and 1000.
+fixed point values between 0 and 1.0 respectively. For a whole slur, their
+int32 values are therefore 0 and 1000.
 
 Parts of the code in this function had to be massaged carefully to ensure that
 exactly the same result is obtained when run native and under valgrind, in
@@ -1027,23 +998,17 @@ Arguments:
   ix1, iy1     coordinates of the end of the slur
   flag         the slur flags
   co           the co ("centre out") value
-  start        the t-value at slur start (fixed point)  ) for drawing
-  stop         the t-value at slur end (fixed point)    )   part-slurs
+  start        the t-value at slur start (fixed point)  ) for drawing partial
+  stop         the t-value at slur end (fixed point)    )   curved slurs
 
 Returns:       nothing
 */
 
 void
-out_slur(int ix0, int iy0, int ix1, int iy1, int flags, int co, int start,
-  int stop)
+out_slur(int32_t ix0, int32_t iy0, int32_t ix1, int32_t iy1, uint32_t flags,
+  int32_t co, int32_t start, int32_t stop)
 {
-int x[10], y[10], cc[10];
-int above, wiggly, ed_adjust;
-double zz[4];
-double temp;
-double xx, yy, w, v;
-double x0, x1, x2, x3, y0, y1, y2, y3;
-double ax, ay, bx, by, cx, cy, f, g;
+int above  = ((flags & sflag_b) == 0)? (+1) : (-1);
 
 if (ix1 == ix0 && iy1 == iy0) return;   /* Avoid crash */
 
@@ -1056,24 +1021,18 @@ if (start == 0 && stop == 1000 && (flags & (sflag_l | sflag_i)) == 0)
   return;
   }
 
-/* Compute values needed by both lines and curved slurs */
-
-xx = (double)ix1 - (double)ix0;
-yy = (double)iy1 - (double)iy0;
-
-above  = ((flags & sflag_b) == 0)? (+1) : (-1);
-
 /* Handle straight-line "slurs". For these, the start and stop values are not
 used, as partial lines are drawn as separate whole lines with appropriate jog
 flags. */
 
 if ((flags & sflag_l) != 0)
   {
-  int lineflags = 0;
-  int adjust = 0;
-  int thickness = (3*main_stavemagn)/10;
-  ix1 += 7*main_stavemagn;
-  co = mac_muldiv((co + 3000)*above, main_stavemagn, 1000);
+  uint32_t lineflags = 0;
+  int32_t adjust = 0;
+  int32_t thickness = (3*out_stavemagn)/10;
+
+  ix1 += 7*out_stavemagn;
+  co = mac_muldiv((co + 3000)*above, out_stavemagn, 1000);
 
   /* Convert the flags to the tie flags used by the ps_line function, then
   output the main portion of the line. Set the savedash flag if drawing a
@@ -1095,7 +1054,7 @@ if ((flags & sflag_l) != 0)
 
   if ((flags & sflag_idot) != 0)
     {
-    thickness = main_stavemagn;
+    thickness = out_stavemagn;
     ps_setdash(out_dashlength, (out_dashgaplength*95)/100, caj_round);
     if (abs(co) < 2*out_dashlength + out_dashgaplength)
       adjust = above*(2*out_dashlength + out_dashgaplength) - co;
@@ -1113,11 +1072,23 @@ if ((flags & sflag_l) != 0)
   return;
   }
 
-/* Compute values needed for curved slur drawing. */
 
-wiggly = ((flags & sflag_w) == 0)? (+1) : (-1);
-w = sqrt(xx*xx + yy*yy)/2.0;
-v = w*0.6667;
+/* Handle the more complicated types of curved slur: partial, dashed, wiggly,
+etc. */
+
+int wiggly = ((flags & sflag_w) == 0)? (+1) : (-1);
+int32_t ed_adjust;
+int32_t x[10], y[10], cc[10];
+
+double zz[4];
+double temp;
+double x0, x1, x2, x3, y0, y1, y2, y3;
+double ax, ay, bx, by, cx, cy, f, g;
+double xx = (double)ix1 - (double)ix0;
+double yy = (double)iy1 - (double)iy0;
+double w = sqrt(xx*xx + yy*yy)/2.0;
+double v = w*0.6667;
+
 if (v > 10000.0) v = 10000.0;
 
 /* It is necessary to use floor() in the conversion of xx*0.3 to an integer in
@@ -1127,7 +1098,7 @@ Using the (int) cast only on a variable (not on a function) avoids a compiler
 warning. */
 
 temp = floor(xx * 0.3);
-co = (above * (co + ((xx > 20000)? 6000 : (int)temp)) * main_stavemagn)/1000;
+co = (above * (co + ((xx > 20000)? 6000 : (int)temp)) * out_stavemagn)/1000;
 
 f = ((double)start)/1000.0;
 g = ((double)stop)/1000.0;
@@ -1138,7 +1109,7 @@ ps_translate, the y value is relative to the stave base. Thereafter use
 ps_abspath() for absolute values. */
 
 ps_gsave();
-ps_translate((ix0+ix1+6*main_stavemagn)/2, (iy0+iy1)/2);
+ps_translate((ix0+ix1+6*out_stavemagn)/2, (iy0+iy1)/2);
 ps_rotate(atan2(yy, xx));
 
 /* Set up traditional Bezier coordinates for the complete slur. */
@@ -1221,7 +1192,8 @@ should be most rare, anyway.) */
 
 if ((flags & sflag_i) != 0)
   {
-  int dashcount, dashlength, gaplength, thickness, length;
+  int dashcount;
+  int32_t dashlength, gaplength, thickness, length;
   double dlength = 0.0;
   double xp = x0;
   double yp = y0;
@@ -1251,8 +1223,8 @@ if ((flags & sflag_i) != 0)
   else
     {
     dashlength = 100;
-    gaplength = ((length < 20000)? 3 : 4) * main_stavemagn;
-    thickness = main_stavemagn;
+    gaplength = ((length < 20000)? 3 : 4) * out_stavemagn;
+    thickness = out_stavemagn;
     }
 
   /* Compute the number of dashes; if greater than one, compute the accurate
@@ -1292,7 +1264,7 @@ coordinates of the control points change. */
 else
   {
   double aay, bby, ccy;
-  ed_adjust = (9*main_stavemagn)/10;
+  ed_adjust = (9*out_stavemagn)/10;
 
   x[4] = x[3];
   x[5] = x[2];
@@ -1300,8 +1272,8 @@ else
   x[7] = x[0];
 
   y0 = -50.0;
-  y1 = (double)(int)(co + above*main_stavemagn + out_slurcly);
-  y2 = (double)(int)(co*wiggly + above*main_stavemagn + out_slurcry);
+  y1 = (double)(int)(co + above*out_stavemagn + out_slurcly);
+  y2 = (double)(int)(co*wiggly + above*out_stavemagn + out_slurcry);
   y3 = -50.0;
 
   aay = y3 - y0 + 3.0*(y1 - y2);
@@ -1332,7 +1304,7 @@ portion. */
 
 if ((flags & sflag_e) != 0 && start < 500 && stop > 500)
   {
-  int xxm, yym;
+  int32_t xxm, yym;
   double dx, dy, theta, cs, sn;
 
   /* Calculate the midpoint of the curve from the parametric equations taking
